#!/usr/bin/env python
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import os
import shutil

import pytest

from sos import execute_workflow


@pytest.mark.skipif(not shutil.which('matlab'), reason='Matlab not installed')
def test_matlab(clear_now_and_after):
    '''Test action matlab'''
    clear_now_and_after('/tmp/matlab_example.txt')
    execute_workflow(r'''
        [0]
        matlab:
            f = fopen("/tmp/matlab_example.txt", "w");
            fprintf(f, "A, B, C, D\n");
            fclose(f);
    ''')
    assert os.path.isfile('/tmp/matlab_example.txt')



#!/usr/bin/env python
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import os
import shutil
import pytest

from sos.targets import file_target
from sos.targets_r import R_library
from sos import execute_workflow


@pytest.mark.skipif(
    not shutil.which("Rscript") or not shutil.which("pandoc"),
    reason="R or pandoc not installed",
)
def test_rmarkdown(clear_now_and_after):
    """Test action Rmarkdown"""
    if not R_library('rmarkdown1').target_exists():
        pytest.xfail('rmarkdown is not properly installed.')

    clear_now_and_after("myreport.html")
    execute_workflow(
        r"""
        [10]

        report:
        ## Some random figure

        Generated by matplotlib


        [100]
        # generate report
        output: 'myreport.html'
        Rmarkdown(output=_output[0])
        """,
        options={"report_output": "report.md"},
    )

    assert os.path.isfile("myreport.html")


@pytest.mark.skipif(
    not shutil.which("Rscript") or not shutil.which("pandoc"),
    reason="R or pandoc not installed",
)
def test_rmarkdown_with_input(clear_now_and_after):
    if not R_library('rmarkdown1').target_exists():
        pytest.xfail('rmarkdown is not properly installed.')

    clear_now_and_after("myreport.html")
    # Rmarkdown with specified input.
    execute_workflow(
        r"""
        [10]
        report: output='a.md'
        ## Some random figure

        Generated by matplotlib


        [100]
        # generate report
        output: 'myreport.html'
        Rmarkdown(input='a.md', output=_output[0])
        """
    )
    assert os.path.isfile("myreport.html")


@pytest.mark.skipif(
    not shutil.which("Rscript") or not shutil.which("pandoc"),
    reason="R or pandoc not installed",
)
def test_rmarkdown_with_action_output(clear_now_and_after):
    if not R_library('rmarkdown1').target_exists():
        pytest.xfail('rmarkdown is not properly installed.')

    clear_now_and_after("default_10.md", "default_20.md", "output.html")
    execute_workflow(
        r"""
        [10]
        report: output='default_10.md'
        A_10

        [20]
        report: output='default_20.md'
        A_20

        [100]
        # generate report
        Rmarkdown(input=['default_10.md', 'default_20.md'], output='output.html')
        """,
        options={"report_output": "${step_name}.md"},
    )
    for f in ["default_10.md", "default_20.md", "output.html"]:
        assert file_target(f).exists()


@pytest.mark.skipif(
    not shutil.which("Rscript") or not shutil.which("pandoc"),
    reason="R or pandoc not installed",
)
def test_rmarkdown_to_stdout():
    if not R_library('rmarkdown1').target_exists():
        pytest.xfail('rmarkdown is not properly installed.')
    execute_workflow(
        r"""
        # generate report
        Rmarkdown:
            # this is title
        """
    )



#!/usr/bin/env python
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import os
import shutil

import pytest

from sos import execute_workflow


@pytest.mark.skipif(not shutil.which('julia'), reason="julia not available")
def test_julia(clear_now_and_after):
    '''Test action Julia'''
    clear_now_and_after('julia_example.txt')
    execute_workflow(r'''
        [0]
        julia:
            open("julia_example.txt", "w") do f
                write(f, "A, B, C, D\n")
            end
    ''')
    assert os.path.isfile('julia_example.txt')



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import subprocess
import textwrap

from sos.converter import extract_workflow


def test_script_to_html(temp_factory):
    '''Test sos show script --html'''
    script1 = textwrap.dedent('''
    [0]
    seq = range(3)
    input: for_each='seq'
    output: "test${_seq}.txt"
    print(output)
    ''')
    script2 = textwrap.dedent('''
    [0]
    seq = range(3)
    input: for_each='seq'
    output: "test${_seq}.txt"
    run:			concurrent=True
        echo 'this is test script'

    [10]
    report('this is action report')
    ''')
    temp_factory('temp1.sos', content=script1)
    temp_factory('temp2.sos', content=script2)

    scripts = ['temp1.sos','temp2.sos']
    for script_file in scripts:
        assert subprocess.call(
            f'sos convert {script_file} {script_file}.html',
            shell=True) == 0
        assert subprocess.call(
            f'sos convert {script_file} {script_file}.html --linenos',
            shell=True) == 0
        #
        assert subprocess.call(['sos', 'convert', script_file, '--to', 'html']) == 0

def test_extract_workflow(sample_workflow):
    '''Test extract workflow from ipynb file'''
    content = extract_workflow('sample_workflow.ipynb')
    print(content)
    assert content == textwrap.dedent('''\
    #!/usr/bin/env sos-runner
    #fileformat=SOS1.0

    # this comment will be included but not shown in help message
    # because it is for the global
    [global]
    a = 1
    # this comment will become the comment for parameter b
    parameter: b=2
    parameter: c=3
    # this comment will become the comment for parameter d
    parameter: d='d'

    # this is a section comment, will be displayed
    [default]
    print(f'Hello {a}')

    ''')



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import multiprocessing as mp
import os
import signal
import time
import pickle
from typing import Any, Dict, Optional

import zmq

from .controller import (close_socket, connect_controllers, create_socket,
                         disconnect_controllers)
from .executor_utils import kill_all_subprocesses, prepare_env
from .utils import env, ProcessKilled, short_repr, get_localhost_ip, get_open_files_and_connections
from .messages import encode_msg, decode_msg


def signal_handler(*args, **kwargs):
    raise ProcessKilled()


class Runner(object):
    '''
    This runner class takea a generator function and run it.
    1. When the generator returns None, continue to run without yielding.
    2. When the generator returns a poller, continue is it receives any message in 0.2s.
    3. Otherwise return False.
    4. The the generator completes, return True.

    So in summary, the Runner returns
    1. True if all completed, return value from generator is ignored.
    2. `self` if waiting.
    '''

    def __init__(self, runner, name):
        self._runner = runner
        self._poller = 0
        self._name = name

    def __repr__(self):
        return self._name

    def run_until_waiting(self):
        try:
            # if poller is not initialized, run the runner to get the first poller
            if self._poller == 0:
                self._poller = next(self._runner)

            while True:
                # if poller is None, meaning we can send None
                # and let the runner continue
                if self._poller is None:
                    self._poller = self._runner.send(None)
                    continue

                # if there is a regular poll, let us wait
                # 0.2 second and see if we can continue
                if self._poller.poll(200):
                    self._poller = self._runner.send(None)
                    continue

                # the poller is not ready, let us break
                return self
        except StopIteration:
            return True

    def can_proceed(self):
        # if there is anything waiting to be continued, proceed.
        return self._poller.poll(0)


class SoS_Worker(mp.Process):
    '''
    Worker process to process SoS step or workflow in separate process. This most
    distinguished feature of this worker is that it allows multiple steps or
    workflows to run, and the worker will yield to other jobs when it needs to
    wait other job to continue. Because all steps and workflows use global varialbes
    in env (e.g. enc.config, enc.sos_dict), we have a stack to store these global
    variables and a mechanism to switch between these contexts.

    The "step" and "subworkflow" are presented as "generators" and are executed by
    so called "runners" that yields if it waits for a socker.

    Each runner is associated with a master socket, which is a PAIR socket. When the
    worker communicates with the worker manager, it sends a list of available "sockets"
    and the worker manager will "claim" the socket when it sends the job to the worker.
    Note that this socket is created by the worker and the other side is supposed to
    connect to it before sending.

    A worker also connects to the master controller, which has two main ports, one for
    push (e.g. signature) to controller, and one for request information from controller.
    These ports are fixed so they are the same for all channels.



    '''

    def __init__(self,
                 config: Optional[Dict[str, Any]] = None,
                 **kwargs) -> None:
        '''

        config:
            values for command line options

            config_file: -c
            output_dag: -d

        args:
            command line argument passed to workflow. However, if a dictionary is passed,
            then it is assumed to be a nested workflow where parameters are made
            immediately available.
        '''
        # the worker process knows configuration file, command line argument etc
        super(SoS_Worker, self).__init__(**kwargs)
        #
        self.config = config
        self.local_ip = get_localhost_ip()

        # there can be multiple jobs for this worker, each using their own port and socket
        self._master_sockets = []
        self._master_ports = []
        # current runner, which can be a runner or True if the runner has completed
        self._runners = []
        # env index, which contains sos_dict for each runner
        self._env_idx = []

    def waiting_runners(self):
        # check if
        return [
            idx for idx, runner in enumerate(self._runners)
            if isinstance(runner, Runner) and runner.can_proceed()
        ]

    def completed_runners(self):
        return [
            idx for idx, runner in enumerate(self._runners) if runner is True
        ]

    def available_ports(self):
        # when a runner is completed, its port becomes available and can
        # be used to accept more jobs.
        return [
            port for port, runner in zip(self._master_ports, self._runners)
            if runner is True
        ]

    def num_pending(self):
        return len(
            [runner for runner in self._runners if isinstance(runner, Runner)])

    def switch_to(self, idx):
        if len(self._master_sockets) > idx:
            # if current stack is ok
            env.master_socket = self._master_sockets[idx]
            env.switch(self._env_idx[idx])
        else:
            assert idx == len(self._master_ports)
            # a new socket is needed
            env.master_socket = create_socket(env.zmq_context, zmq.PAIR)
            port = env.master_socket.bind_to_random_port(
                f'tcp://{self.local_ip}')
            # switch to a new env_idx and returns new_idx, old_idx
            self._env_idx.append(env.request_new()[0])
            self._master_sockets.append(env.master_socket)
            self._master_ports.append(f'tcp://{self.local_ip}:{port}')
            self._runners.append(True)
            if 'WORKER' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
                    'SOS_DEBUG']:
                env.log_to_file(
                    'WORKER',
                    f'WORKER {self.name} ({os.getpid()}) creates ports {self._master_ports}'
                )

    def __repr__(self):
        return self.name + ' ' + ' '.join(
            str(x) if isinstance(x, Runner) else str(idx)
            for idx, x in enumerate(self._runners))

    def run(self):
        # env.logger.warning(f'Worker created {os.getpid()}')
        env.config.update(self.config)
        env.verbosity = self.config.get('verbosity', 2)

        if 'PROFILE' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
                'SOS_DEBUG']:
            import cProfile
            pr = cProfile.Profile()
            pr.enable()

        env.zmq_context = connect_controllers()

        # create controller socket
        env.ctrl_socket = create_socket(env.zmq_context, zmq.REQ,
                                        'worker backend')
        # worker_backend, or the router, might be on another machine
        env.log_to_file(
            'WORKER',
            f'Connecting to router {self.config["sockets"]["worker_backend"]} from {get_localhost_ip()}'
        )
        env.ctrl_socket.connect(self.config["sockets"]["worker_backend"])

        signal.signal(signal.SIGTERM, signal_handler)
        # result socket used by substeps
        env.result_socket = None
        env.result_socket_port = None

        # wait to handle jobs
        while True:
            try:
                if 'OPENFILES' in env.config[
                        'SOS_DEBUG'] or 'ALL' in env.config['SOS_DEBUG']:
                    ofc = get_open_files_and_connections(os.getpid())
                    env.log_to_file(
                        'OPENFILES',
                        f'WORKER {os.getpid()} has {len(ofc["open_files"])} open files and {len(ofc["connections"])} connections.'
                    )
                wr = self.waiting_runners()
                if wr:
                    for idx in wr:
                        self.switch_to(idx)
                        # it can be True for completion and Runner itself for continue
                        self._runners[idx] = self._runners[
                            idx].run_until_waiting()
                    continue

                cr = self.completed_runners()
                # using an completed slot or create a new one
                new_idx = len(self._runners) if not cr else cr[0]
                self.switch_to(new_idx)

                # although we have chosen one port, but we hae advertised multiple ports
                # and the executor might choose another one. We therefore need to send all
                # avilable ports to the controller. We also need to send a flag to let the
                # controller know if we have any pending job, and the controller might decide
                # to kill this worker.
                env.ctrl_socket.send(
                    encode_msg([self.num_pending()] + self.available_ports()))
                reply = decode_msg(env.ctrl_socket.recv())

                if reply is None:
                    if len(wr) != 0:
                        env.logger.error(
                            f'WORKER terminates with pending tasks. sos might not be termianting properly.'
                        )
                    if 'WORKER' in env.config[
                            'SOS_DEBUG'] or 'ALL' in env.config['SOS_DEBUG']:
                        env.log_to_file(
                            'WORKER',
                            f'WORKER {self.name} ({os.getpid()}) quits after receiving None.'
                        )
                    break
                if not reply:  # if an empty job is returned
                    time.sleep(0.1)
                    continue

                #
                # if a real job is returned, run it. _process_job will either return True
                # or a runner in case it is interrupted.
                env.log_to_file(
                    'WORKER',
                    f'WORKER {self.name} ({os.getpid()}, {self.num_pending()} pending) receives {self._type_of_work(reply)} request {self._name_of_work(reply)} with master port {self._master_ports[new_idx]}'
                )

                if 'task' in reply:
                    self.run_substep(reply)
                    env.log_to_file(
                        'WORKER',
                        f'WORKER {self.name} ({os.getpid()}) completes substep {self._name_of_work(reply)}'
                    )
                    self._runners[new_idx] = True
                    continue
                elif 'task_id' in reply:
                    self.run_task(reply)
                    env.log_to_file(
                        'WORKER',
                        f'WORKER {self.name} ({os.getpid()}) completes task {self._name_of_work(reply)}'
                    )
                    self._runners[new_idx] = True
                    continue

                master_port = reply['config']['sockets']['master_port']
                if master_port != self._master_ports[new_idx]:
                    new_idx = self._master_ports.index(master_port)
                    self.switch_to(new_idx)

                # step and workflow can yield. Here we call run_until_waiting directly because we know the Runner can proceed.
                self._runners[new_idx] = Runner(
                    self.run_step(**reply)
                    if 'section' in reply else self.run_workflow(**reply),
                    name=self._name_of_work(reply)).run_until_waiting()
                if 'WORKER' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
                        'SOS_DEBUG']:
                    env.log_to_file(
                        'WORKER',
                        'STATUS ' + self._name_of_work(reply) + str(self))
            except ProcessKilled:
                # in theory, this will not be executed because the exception
                # will be caught by the step executor, and then sent to the master
                # process, which will then trigger terminate() and send a None here.
                break
            except KeyboardInterrupt:
                # ignore SIGINT because workers are supposed to be killed through messages
                env.log_to_file(
                    'PROCESS',
                    f'KeyboardInterrupt received by {os.getpid()}. Ignoring.')
        # Finished
        signal.signal(signal.SIGTERM, signal.SIG_DFL)
        kill_all_subprocesses(os.getpid())

        close_socket(env.result_socket, 'substep result', now=True)

        for socket in self._master_sockets:
            close_socket(socket, 'worker master', now=True)
        close_socket(env.ctrl_socket, now=True)
        disconnect_controllers(env.zmq_context)
        if 'PROFILE' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
                'SOS_DEBUG']:
            pr.disable()
            pr_file = os.path.join(
                os.path.expanduser('~'), '.sos',
                f'profile_worker_{os.getpid()}.txt')
            pr.dump_stats(pr_file)
            print(
                f'Execution profile of worker process {os.getpid()} is saved to {pr_file}'
            )

    def _type_of_work(self, work):
        if 'section' in work:
            return 'step'
        elif 'wf' in work:
            return 'workflow'
        elif 'task_id' in work:
            return 'task'
        else:
            return 'substep'

    def _name_of_work(self, work):
        if 'section' in work:
            return work['section'].step_name()
        elif 'wf' in work:
            return work['workflow_id']
        elif 'task_id' in work:
            return work['task_id']
        else:
            return 'substep'

    def run_workflow(self, workflow_id, wf, targets, args, shared, config,
                     **kwargs):
        #
        #
        # get workflow, args, shared, and config
        from .workflow_executor import Base_Executor
        env.config.update(config)
        # we are in a separate process and need to set verbosity from workflow config
        # but some tests do not provide verbosity
        env.verbosity = config.get('verbosity', 2)
        env.log_to_file(
            'WORKER',
            f'Worker {self.name} working on a workflow {workflow_id} with args {short_repr(args)}'
        )

        executer = Base_Executor(wf, args=args, shared=shared, config=config)
        # we send the socket to subworkflow, which would send
        # everything directly to the master process, so we do not
        # have to collect result here
        try:
            runner = executer.run_as_nested(
                targets=targets,
                parent_socket=env.master_socket,
                my_workflow_id=workflow_id)
            try:
                yreq = next(runner)
                while True:
                    yres = yield yreq
                    yreq = runner.send(yres)
            except StopIteration:
                pass

        except Exception as e:
            env.master_socket.send(encode_msg(e))

    def run_step(self, section, context, shared, args, config, verbosity):
        from .step_executor import Step_Executor

        env.log_to_file(
            'WORKER',
            f'Worker {self.name} working on {section.step_name()} with args {short_repr(args)}'
        )
        env.config.update(config)
        env.verbosity = verbosity
        #
        # Execute global namespace. The reason why this is executed outside of
        # step is that the content of the dictioary might be overridden by context
        # variables.
        prepare_env(section.global_def, section.global_vars,
                    env.config['workflow_vars'])

        # clear existing keys, otherwise the results from some random result
        # might mess with the execution of another step that does not define input
        for k in ['__step_input__', '__default_output__', '__step_output__']:
            if k in env.sos_dict:
                env.sos_dict.pop(k)
        # if the step has its own context
        env.sos_dict.quick_update(shared)
        # context should be updated after shared because context would contain the
        # correct __step_output__ of the step, whereas shared might contain
        # __step_output__ from auxiliary steps. #526
        env.sos_dict.quick_update(context)
        executor = Step_Executor(
            section, env.master_socket, mode=env.config['run_mode'])

        runner = executor.run()
        try:
            yreq = next(runner)
            while True:
                yres = yield yreq
                yreq = runner.send(yres)
        except StopIteration:
            pass

    def run_substep(self, work):
        from .substep_executor import execute_substep
        execute_substep(**work)

    def run_task(self, work):
        from .task_executor import BaseTaskExecutor
        executor = BaseTaskExecutor()

        result_socket = work['config']['sockets']['result_push_socket']

        if env.result_socket_port is not None and env.result_socket_port != result_socket:
            close_socket(env.result_socket)
            env.result_socket = None

        if env.result_socket is None:
            env.result_socket = create_socket(env.zmq_context, zmq.PUSH)
            env.result_socket_port = result_socket
            # the result_socket_port contains IP of the worker that request the substep
            env.result_socket.connect(env.result_socket_port)

        res = executor.execute_single_task(**work)
        env.result_socket.send(encode_msg(res))


class WorkerManager(object):
    # manager worker processes

    def __init__(self, worker_procs, backend_socket):
        if isinstance(worker_procs, (int, str)):
            self._worker_procs = [str(worker_procs)]
        elif worker_procs is None:
            # from sos notebook with no value specified
            self._worker_procs = [str(min(max(os.cpu_count() // 2, 2), 8))]
        else:
            # should be a sequence
            self._worker_procs = worker_procs

        # the first item in self._worker_procs is always considered to be the localhost, which is where
        # the router lives. The rest of the hosts will be considered as remote workers.
        try:

            self._worker_hosts = []
            self._max_workers = []
            for worker_proc in self._worker_procs:
                if ':' in worker_proc:
                    worker_host, max_workers = worker_proc.rsplit(':', 1)
                    if not max_workers.isdigit():
                        raise ValueError(
                            f'Invalid worker specification {worker_proc}: number of process expected after ":"'
                        )
                    self._worker_hosts.append(worker_host)
                    self._max_workers.append(int(max_workers))
                elif worker_proc.isdigit():
                    self._worker_hosts.append(get_localhost_ip())
                    self._max_workers.append(int(worker_proc))
                else:
                    self._worker_hosts.append(worker_proc)
                    # here we assume that all nodes have the same number of cores so that we use
                    # the default value for master node for all computing nodes
                    self._max_workers.append(
                        min(max(os.cpu_count() // 2, 2), 8))

            self._num_workers = [0 for x in self._worker_procs]
        except Exception:
            raise RuntimeError(
                f'Incorrect format for option -j ({self._worker_procs}), which should be one or more [host:]nproc'
            )

        self._local_workers = []
        self._remote_connections = []

        self._num_remote_workers = {}

        self._n_requested = 0
        self._n_processed = 0

        self._local_worker_alive_time = time.time()
        # self._last_pending_time = {}

        self._substep_requests = []
        self._task_requests = []
        self._step_requests = {}

        self._worker_backend_socket = backend_socket

        # ports of workers working for blocking workflow
        self._blocking_ports = set()

        self._available_ports = set()
        self._claimed_ports = set()

        self._last_pending_msg = {}

        # start a worker, note that we do not start all workers for performance
        # considerations
        self.start_worker()

    def report(self, msg):
        if 'WORKER' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
                'SOS_DEBUG']:
            env.log_to_file(
                'WORKER',
                f'{msg.upper()}: {self._num_workers} workers (of which {len(self._blocking_ports)} is blocking), {self._n_requested} requested, {self._n_processed} processed'
            )

    def add_request(self, msg_type, msg):
        self._n_requested += 1
        if msg_type == 'substep':
            self._substep_requests.insert(0, msg)
            self.report(f'Substep requested')
        elif msg_type == 'task':
            self._task_requests.insert(0, msg)
            self.report(f'Task requested')
        else:
            port = msg['config']['sockets']['master_port']
            self._step_requests[port] = msg
            self.report(f'Step {port} requested')

        if sum(self._num_workers) < sum(self._max_workers):
            self.start_worker()

    def worker_available(self, blocking, excluded):
        if self._available_ports:
            usable = [x for x in self._available_ports if x not in excluded]
            if usable:
                claimed = usable[0]
                self._available_ports.remove(usable[0])
                self._claimed_ports.add(claimed)
                return claimed

        if not blocking:
            # no available port, can we start a new worker?
            if sum(self._num_workers) < sum(self._max_workers):
                self.start_worker()
            return None

        # we start a worker right now.
        self._max_workers[0] += 1
        self.start_worker()
        while True:
            if not self._worker_backend_socket.poll(5000):
                raise RuntimeError('No worker is started after 5 seconds')
            msg = decode_msg(self._worker_backend_socket.recv())
            port = self.process_request(msg[0], msg[1:], request_blocking=True)
            if port is None or port in excluded:
                continue
            self._claimed_ports.add(port)
            self._blocking_ports.add(port)
            env.logger.debug(
                f'Increasing maximum number of local workers to {self._max_workers[0]} to accommodate a blocking subworkflow.'
            )
            return port

    def process_request(self, num_pending, ports, request_blocking=False):
        '''port is the open port at the worker, num_pending is the num_pending of stack.
        A non-zero num_pending means that the worker is pending on something while
        looking for new job, so the worker should not be killed.
        '''
        if any(port in self._step_requests for port in ports):
            # if the port is available
            port = [x for x in ports if x in self._step_requests][0]
            self._worker_backend_socket.send(
                encode_msg(self._step_requests.pop(port)))
            self._n_processed += 1
            self.report(f'Step {port} processed')
            # port should be in claimed ports
            self._claimed_ports.remove(port)
            # if ports[0] in self._last_pending_time:
            #     self._last_pending_time.pop(ports[0])
        elif any(port in self._claimed_ports for port in ports):
            # the port is claimed, but the real message is not yet available
            self._worker_backend_socket.send(encode_msg({}))
            self.report(f'pending with claimed {ports}')
        # elif any(port in self._blocking_ports for port in ports):
        #     # in block list but appear to be idle, kill it
        #     self._max_workers -= 1
        #     env.logger.debug(
        #         f'Reduce maximum number of workers to {self._max_workers} after completion of a blocking subworkflow.'
        #     )
        #     for port in ports:
        #         if port in self._blocking_ports:
        #             self._blocking_ports.remove(port)
        #         if port in self._available_ports:
        #             self._available_ports.remove(port)
        #     self._worker_backend_socket.send(encode_msg(None))
        #     self._num_local_workers -= 1
        #     self.report(f'Blocking worker {ports} killed')
        elif self._task_requests:
            # port is not claimed, free to use for substep worker
            msg = self._task_requests.pop()
            self._worker_backend_socket.send(encode_msg(msg))
            self._n_processed += 1
            self.report(f'Task processed with {ports[0]}')
            # port can however be in available ports
            for port in ports:
                if port in self._available_ports:
                    self._available_ports.remove(port)
                # if port in self._last_pending_time:
                #     self._last_pending_time.pop(port)
        elif self._substep_requests:
            # port is not claimed, free to use for substep worker
            msg = self._substep_requests.pop()
            self._worker_backend_socket.send(encode_msg(msg))
            self._n_processed += 1
            self.report(f'Substep processed with {ports[0]}')
            # port can however be in available ports
            for port in ports:
                if port in self._available_ports:
                    self._available_ports.remove(port)
                # if port in self._last_pending_time:
                #     self._last_pending_time.pop(port)
        elif request_blocking:
            self._worker_backend_socket.send(encode_msg({}))
            return ports[0]
        # elif num_pending == 0 and self._num_local_workers > 1 and ports[
        #         0] in self._last_pending_time and time.time(
        #         ) - self._last_pending_time[ports[0]] > 5:
        #     # kill the worker
        #     for port in ports:
        #         if port in self._available_ports:
        #             self._available_ports.remove(port)
        #     self._worker_backend_socket.send(encode_msg(None))
        #     self._num_local_workers -= 1
        #     self.report(f'Kill standing {ports}')
        #     self._last_pending_time.pop(ports[0])
        else:
            # if num_pending == 0 and ports[0] not in self._last_pending_time:
            #     self._last_pending_time[ports[0]] = time.time()
            self._available_ports.add(ports[0])
            self._worker_backend_socket.send(encode_msg({}))
            ports = tuple(ports)
            if (ports, num_pending) not in self._last_pending_msg or time.time(
            ) - self._last_pending_msg[(ports, num_pending)] > 1.0:
                self.report(
                    f'pending with port {ports} at num_pending {num_pending}')
                self._last_pending_msg[(ports, num_pending)] = time.time()

    def start_worker(self):
        for idx, (worker_host, num_worker, max_worker) in enumerate(
                zip(self._worker_hosts, self._num_workers, self._max_workers)):
            if num_worker == max_worker:
                continue
            # local host
            if idx == 0:
                worker = SoS_Worker(env.config)
                worker.start()
                self._local_worker_alive_time = time.time()
                self._local_workers.append(worker)
                self._num_workers[0] += 1
                self.report('start a local worker')
            else:
                # start all remote workers on a host
                try:
                    from .hosts import Host
                    host = Host(worker_host, start_engine=False)
                    env_file = os.path.join(
                        os.path.expanduser('~'), '.sos',
                        f'worker_envs.{os.getpid()}')
                    if not os.path.isfile(env_file):
                        with open(env_file, 'wb') as envfile:
                            pickle.dump(
                                {
                                    x: y
                                    for x, y in os.environ.items()
                                    if isinstance(y, str)
                                }, envfile)
                    # NOTE: we assume file systems are shared so we do not copy env_file to remote host
                    cmd = [
                        'sos', 'worker', '--router',
                        env.config["sockets"]["worker_backend"], '--sig_mode',
                        env.config['sig_mode'], '--run_mode',
                        env.config['run_mode'], '--env', env_file, '--workdir',
                        os.getcwd(), '-v',
                        str(env.config['verbosity'])
                    ]
                    if max_worker is not None:
                        cmd += ['-j', str(max_worker)]
                    p = host._host_agent.run_command(cmd, wait_for_task=False)
                    self._remote_connections.append(p)
                except Exception as e:
                    env.logger.error(
                        f'Failed to start workers on host {worker_host}: {e}')
                    raise RuntimeError(
                        f'Failed to start workers on host {worker_host}: {e}')
                self._num_workers[idx] = self._max_workers[idx]
                self.report(
                    f'start {max_worker} remote workers on {worker_host}')
            break

    def check_workers(self):
        '''Kill workers that have been pending for a while and check if all workers
        are alive. '''
        if time.time() - self._local_worker_alive_time > 5:
            self._local_worker_alive_time = time.time()
            # join processes if they are now gone, it should not do anything bad
            # if the process is still running
            [
                worker.join()
                for worker in self._local_workers
                if not worker.is_alive()
            ]
            self._local_workers = [
                worker for worker in self._local_workers if worker.is_alive()
            ]
            if len(self._local_workers) < self._num_workers[0]:
                raise ProcessKilled('One of the local workers has been killed.')

    def kill_all(self):
        '''Kill all workers'''
        total_num_workers = sum(self._num_workers)
        while total_num_workers > 0 and self._worker_backend_socket.poll(1000):
            msg = decode_msg(self._worker_backend_socket.recv())
            self._worker_backend_socket.send(encode_msg(None))
            total_num_workers -= 1
            self.report(f'Kill {msg[1:]}')
        # join all local processes
        [worker.join() for worker in self._local_workers]



#!/usr/bin/env python
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import shutil

import pytest

from sos import execute_workflow


@pytest.mark.skipif(not shutil.which('node'), reason='node not installed')
def test_node():
    '''Test action node'''
    execute_workflow(r'''
        [0]
        node:
        var args = process.argv.slice(2);
        console.log('Hello ' + args.join(' ') + '!');
        ''')



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import os
import pytest
import shutil
import subprocess
import sys
import time
import unittest
import sqlite3
from contextlib import contextmanager

from sos.hosts import Host
from sos.parser import SoS_Script
from sos.targets import file_target
from sos.utils import env
from sos.tasks import TaskParams, TaskFile
from sos import execute_workflow
from sos.workflow_executor import Base_Executor

has_docker = sys.platform != "win32"
try:
    if sys.platform != "win32":
        subprocess.check_output("docker ps | grep test_sos", shell=True).decode()
except subprocess.CalledProcessError:
    subprocess.call("sh build_test_docker.sh", shell=True)
    try:
        subprocess.check_output("docker ps | grep test_sos", shell=True).decode()
    except subprocess.CalledProcessError:
        print("Failed to set up a docker machine with sos")
        has_docker = False


@contextmanager
def cd_new(path):
    old_dir = os.getcwd()
    if os.path.isdir(path):
        shutil.rmtree(path)
    os.makedirs(path, exist_ok=True)
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(old_dir)


def get_tasks():
    conn = sqlite3.connect(".sos/workflow_signatures.db")
    cur = conn.cursor()
    cur.execute('SELECT DISTINCT id FROM workflows WHERE entry_type = "task"')
    return [x[0] for x in cur.fetchall()]


class TestTask(unittest.TestCase):
    def setUp(self):
        env.reset()
        subprocess.call("sos remove -s", shell=True)
        # self.resetDir('~/.sos')
        self.temp_files = []
        Host.reset()

    def touch(self, files):
        """create temporary files"""
        if isinstance(files, str):
            files = [files]
        #
        for f in files:
            with open(f, "w") as tmp:
                tmp.write("test")
        #
        self.temp_files.extend(files)

    def test_task_file(self):
        """Test task file handling"""
        for ext in (".pulse", ".out", ".err", ".task", ".sh"):
            filename = os.path.join(
                os.path.expanduser("~"), ".sos", "tasks", "ffffffffffffffff" + ext
            )
            if os.path.isfile(filename):
                os.remove(filename)
        params = TaskParams(
            name="ffffffffffffffff",
            global_def={},
            task="b=a",
            sos_dict={"a": 1},
            tags=["b", "a"],
        )
        a = TaskFile("ffffffffffffffff")
        a.save(params)
        self.assertEqual(a.tags, "a b")
        for ext in (".pulse", ".out", ".err", ".sh"):
            with open(
                os.path.join(
                    os.path.expanduser("~"), ".sos", "tasks", "ffffffffffffffff" + ext
                ),
                "w",
            ) as fh:
                fh.write(ext)
        self.assertFalse(a.has_stdout())
        self.assertFalse(a.has_stderr())
        a.add_outputs()
        #
        self.assertEqual(a.params.sos_dict["a"], 1)
        self.assertEqual(a.status, "new")
        a.status = "completed"
        self.assertLess(time.time() - a.last_updated, 2)
        self.assertEqual(a.status, "completed")
        #
        # get and reset info
        info = a.info
        a.status = "running"
        self.assertEqual(a.status, "running")
        a.info = info
        self.assertEqual(a.status, "completed")
        self.assertTrue(a.has_stdout())
        #
        a.add_result({"ret_code": 5})
        #
        a.tags = ["ee", "d"]
        self.assertEqual(a.tags, "d ee")
        # a.add_tags(['kk'])
        # self.assertEqual(a.tags.split(), ['d', 'ee', 'kk'])
        #
        self.assertEqual(a.params.sos_dict["a"], 1)
        self.assertEqual(a.params.task, "b=a")
        #

        self.assertEqual(a.stdout, ".out")
        self.assertEqual(a.stderr, ".err")
        self.assertEqual(a.pulse, ".pulse")
        self.assertEqual(a.shell, ".sh")
        self.assertTrue(a.has_stdout())
        self.assertTrue(a.has_stderr())
        self.assertTrue(a.has_pulse())
        self.assertTrue(a.has_shell())
        #
        #
        a.reset()
        self.assertEqual(a.status, "new")
        self.assertEqual(a.stdout, "")
        self.assertEqual(a.stderr, "")
        self.assertEqual(a.signature, {})
        a.add_outputs()
        a.add_result({"ret_code": 5})
        self.assertEqual(a.result["ret_code"], 5)

    def test_workdir(self):
        """Test workdir option for runtime environment"""
        import tempfile

        tdir = tempfile.mkdtemp()
        with open(os.path.join(tdir, "aaa.pp"), "w") as aaa:
            aaa.write("something")
        script = r"""
import os
[0]
task: workdir={0!r}

with open(os.path.join({1!r}, 'result.txt'), 'w') as res:
   for file in os.listdir({1!r}):
       res.write(file + '\n')
""".format(
            os.path.split(tdir)[0], os.path.split(tdir)[1]
        )
        wf = SoS_Script(script).workflow()
        env.config["sig_mode"] = "force"
        Base_Executor(wf, config={"default_queue": "localhost"}).run()
        with open(os.path.join(tdir, "result.txt")) as res:
            content = [x.strip() for x in res.readlines()]
            self.assertTrue("aaa.pp" in content)

    #     def testSequential(self):
    #         '''Test concurrency option for runtime environment'''
    #         env.max_jobs = 5
    #         start = time.time()
    #         execute_workflow(
    #             script=r"""
    # import time
    # [0]

    # repeat = range(4)
    # input: for_each='repeat'

    # task: concurrent=False

    # print('I am {}, waited {} seconds'.format(_index, _repeat + 1))
    # time.sleep(_repeat + 1)
    # print('I am {}, done'.format(_index))
    # """,
    #             options={
    #                 'default_queue': 'localhost',
    #                 'sig_mode': 'force'
    #             })
    #         self.assertGreater(time.time() - start, 11)

    def test_concurrency(self):
        """Test concurrency option for runtime environment"""
        execute_workflow(
            r"""
[0]

repeat = range(4)
input: for_each='repeat'

task:

import time
print('I am {}, waited {} seconds'.format(_index, _repeat + 1))
time.sleep(_repeat + 1)
print('I am {}, done'.format(_index))
""",
            options={"default_queue": "localhost", "sig_mode": "force"},
        )

    def test_prepend_path(self):
        """Test prepend path"""
        import stat

        if not os.path.isdir("temp"):
            os.mkdir("temp")
        if sys.platform == "win32":
            with open(r"temp\temp_cmd.bat", "w") as tc:
                tc.write('echo "a"')
        else:
            with open("temp/temp_cmd", "w") as tc:
                tc.write('echo "a"')
            os.chmod("temp/temp_cmd", stat.S_IXUSR | stat.S_IWUSR | stat.S_IRUSR)
        #
        script = SoS_Script(
            r"""
[1]
task:
run:
    temp_cmd
"""
        )
        script.workflow()
        env.config["sig_mode"] = "force"
        # self.assertRaises(Exception, Base_Executor(wf).run)
        #
        # the following is supposed to create its own task file but
        # for some reason it uses the same task file
        #
        # use option env
        execute_workflow(
            script=r"""
import os
[1]
task: env={'PATH': 'temp' + os.pathsep + os.environ['PATH']}
run:
    temp_cmd
""",
            options={"default_queue": "localhost", "sig_mode": "force"},
        )
        #
        #
        execute_workflow(
            script=r"""
[1]
task: prepend_path='temp'
run:
    temp_cmd
""",
            options={"default_queue": "localhost"},
        )

    def test_no_task(self):
        env.config["sig_mode"] = "force"
        script = SoS_Script(
            r"""

[10]
task:
run:
   sleep 0
"""
        )
        wf = script.workflow()
        # this will always work and not through task
        Base_Executor(wf, config={"default_queue": None}).run()
        #
        env.config["sig_mode"] = "force"
        script = SoS_Script(
            r"""

[10]
task: queue=None
run:
   sleep 0
"""
        )
        wf = script.workflow()
        # this will always work and not through task
        Base_Executor(wf, config={"default_queue": "localhost"}).run()

    def test_passing_var_to_task(self):
        """Test passing used variable to tasks"""
        for i in range(10, 13):
            if file_target(f"myfile_{i}.txt").exists():
                file_target(f"myfile_{i}.txt").unlink()
        #
        env.config["sig_mode"] = "force"
        script = SoS_Script(
            r"""
parameter: gvar = 10

[10]
# generate a file
tt = range(gvar, gvar + 3)
input: for_each='tt'
output: f"myfile_{_tt}.txt"
# additional comment

# _tt should be used in task
task:
run: expand=True
    echo {_tt}_{_index} > {_output:q}

"""
        )
        wf = script.workflow()
        env.max_jobs = 4
        Base_Executor(wf, config={"default_queue": "localhost"}).run()
        for t in range(10, 13):
            with open(f"myfile_{t}.txt") as tmp:
                self.assertEqual(tmp.read().strip(), str(t) + "_" + str(t - 10))
            if file_target(f"myfile_{t}.txt").exists():
                file_target(f"myfile_{t}.txt").unlink()

    def test_max_jobs(self):
        """Test default max number of jobs"""
        script = SoS_Script(
            r"""

[10]
input: for_each=[{'a': range(2)}, {'b': range(3)}]

task:
run: expand=True
    echo "a = {a}, b = {b}"
"""
        )
        env.config["max_running_jobs"] = 2
        wf = script.workflow()
        Base_Executor(wf, config={"default_queue": "localhost"}).run()

    def test_kill_and_purge(self):
        """Test no wait"""
        subprocess.call(["sos", "purge"])
        with open("test_purge.sos", "w") as script:
            script.write(
                r"""
[10]
task:
run:
    echo Try to kill
    sleep 20
"""
            )
        subprocess.Popen("sos run test_purge.sos -s force -q localhost", shell=True)
        time.sleep(5)
        subprocess.call(["sos", "kill", "--all"])
        for i in range(20):
            output = subprocess.check_output(["sos", "status", "-v", "1"]).decode()
            if "killed" in output or "aborted" in output or "completed" in output:
                break
            self.assertFalse(
                i > 10, "Task should be killed within 10 seconds, got {}".format(output)
            )
            time.sleep(1)
        # test purge by status
        subprocess.call(["sos", "purge", "--status", "aborted"])
        self.assertFalse(
            "killed" in subprocess.check_output(["sos", "status", "-v", "3"]).decode()
        )
        # purge by all is not tested because it is dangerous

    def test_concurrent_task(self):
        """Test submitting tasks from concurrent substeps"""
        for f in [f"con_{x}.txt" for x in range(5)]:
            if file_target(f).exists():
                file_target(f).unlink()
        script = SoS_Script(
            """
[10]
input: for_each={'i': range(5)}
output: f'con_{i}.txt'

task:
run: expand=True
  echo {i} > {_output}
"""
        )
        wf = script.workflow()
        Base_Executor(
            wf, config={"sig_mode": "force", "default_queue": "localhost"}
        ).run()
        for f in [f"con_{x}.txt" for x in range(5)]:
            self.assertTrue(file_target(f).exists())

    def test_shared_option(self):
        """Test shared option of task"""
        for f in ("a.txt", "a100.txt"):
            if file_target(f).exists():
                file_target(f).unlink()
        script = SoS_Script(
            """
[10: shared = 'a']
output: 'a.txt'
task: shared={'a': 'int(open("a.txt").read())'}
run:
  echo 100 > a.txt

[20]
run: expand=True
    touch a{a}.txt
"""
        )
        wf = script.workflow()
        Base_Executor(
            wf, config={"sig_mode": "force", "default_queue": "localhost"}
        ).run()
        self.assertTrue(os.path.isfile("a100.txt"))
        # sequence of var or mapping
        for f in ("a.txt", "a100.txt"):
            if file_target(f).exists():
                file_target(f).unlink()
        script = SoS_Script(
            """
[10: shared = ['a', 'b']]
output: 'a.txt'
task: shared=[{'a': 'int(open("a.txt").read())'}, 'b']
b = 20
run:
  echo 100 > a.txt

[20]
run: expand=True
    touch a{a}_{b}.txt
"""
        )
        wf = script.workflow()
        Base_Executor(
            wf, config={"sig_mode": "force", "default_queue": "localhost"}
        ).run()
        self.assertTrue(os.path.isfile("a100_20.txt"))

        script = SoS_Script(
            """
[10 (simulate): shared=['rng', 'step_rng']]
input: for_each={'i': range(5)}
task: shared='rng'
print(f"{i}")
import random
rng = random.randint(1, 1000)
"""
        )
        wf = script.workflow()
        Base_Executor(wf, config={"default_queue": "localhost"}).run()
        var = env.sos_dict["rng"]
        self.assertTrue(isinstance(var, int))
        self.assertTrue(isinstance(env.sos_dict["step_rng"], list))
        self.assertEqual(env.sos_dict["step_rng"][-1], var)
        # run it again, should get from signature
        #
        # Base_Executor(wf).run()
        # self.assertEqual(var, env.sos_dict['rng'])

    def test_trunk_size_option(self):
        """Test option trunk_size"""
        with open("test_trunksize.sos", "w") as tt:
            tt.write(
                """
[10]
input: for_each={'I': range(10)}
task: trunk_size=5, cores=1, mem='1M', walltime='10m'
run: expand=True
    echo {I} > {I}.txt
    sleep 0.1
"""
            )
        wf = SoS_Script(filename="test_trunksize.sos").workflow()
        for i in range(10):
            if os.path.isfile(f"{i}.txt"):
                file_target(f"{i}.txt").unlink()
        Base_Executor(
            wf,
            config={
                "sig_mode": "force",
                "script": "test_trunksize.sos",
                "max_running_jobs": 10,
                "workflow_args": [],
                "output_dag": "",
                "output_report": None,
                "targets": [],
                "worker_procs": ["4"],
                "default_queue": "localhost",
                "workflow": "default",
                "workdir": ".",
            },
        ).run()

        for i in range(10):
            self.assertTrue(os.path.isfile(f"{i}.txt"))
            os.remove(f"{i}.txt")
        # trunk size is None or 0, -1, intepreted as all tasks
        with open("test_trunksize.sos", "w") as tt:
            tt.write(
                """
[10]
input: for_each={'I': range(10)}
task: trunk_size=None, cores=1, mem='1M', walltime='10m'
run: expand=True
    echo {I} > {I}.txt
    sleep 0.1
"""
            )
        wf = SoS_Script(filename="test_trunksize.sos").workflow()
        Base_Executor(
            wf,
            config={
                "sig_mode": "force",
                "script": "test_trunksize.sos",
                "max_running_jobs": 10,
                "workflow_args": [],
                "output_dag": "",
                "output_report": None,
                "targets": [],
                "worker_procs": ["4"],
                "default_queue": "localhost",
                "workflow": "default",
                "workdir": ".",
            },
        ).run()
        for i in range(10):
            self.assertTrue(os.path.isfile(f"{i}.txt"))

    def test_trunk_workers_option(self):
        """Test option trunk_workers"""
        with open("test_trunkworker.sos", "w") as tt:
            tt.write(
                """
[10]
input: for_each={'I': range(12)}
task: trunk_size=6, trunk_workers=3, mem='1M', walltime='10m'
run: expand=True
    echo {I} > {I}.txt
    sleep 2
"""
            )
        wf = SoS_Script(filename="test_trunkworker.sos").workflow()
        Base_Executor(
            wf,
            config={
                "sig_mode": "force",
                "script": "test_trunkworker.sos",
                "max_running_jobs": 10,
                "workflow_args": [],
                "output_dag": "",
                "output_report": None,
                "targets": [],
                "worker_procs": ["4"],
                "default_queue": "localhost",
                "workflow": "default",
                "workdir": ".",
            },
        ).run()
        for i in range(10):
            self.assertTrue(os.path.isfile("{}.txt".format(i)))

    def test_task_tags(self):
        """Test option tags of tasks"""
        import random

        tag = "tag{}".format(random.randint(1, 100000))
        with open("test_tags.sos", "w") as tt:
            tt.write(
                """
[10]
input: for_each={{'i': range(10)}}
task: tags='{}', trunk_size=2
sh: expand=True
  echo {} {{i}}
""".format(
                    tag, tag
                )
            )
        wf = SoS_Script(filename="test_tags.sos").workflow()
        Base_Executor(
            wf,
            config={
                "sig_mode": "force",
                "script": "test_trunkworker.sos",
                "max_running_jobs": 10,
                "workflow_args": [],
                "output_dag": "",
                "output_report": None,
                "default_queue": "localhost",
                "targets": [],
                "worker_procs": ["4"],
                "workflow": "default",
                "workdir": ".",
            },
        ).run()
        ret = subprocess.check_output(
            "sos status -t {}".format(tag), shell=True
        ).decode()
        self.assertEqual(len(ret.splitlines()), 5, "Obtained {}".format(ret))
        # test multiple tags
        tag1 = "tag{}".format(random.randint(1, 100000))
        tag2 = "tag{}".format(random.randint(1, 100000))
        with open("test_tags.sos", "w") as tt:
            tt.write(
                """
[10]
input: for_each={{'i': range(2)}}
task: tags=['{}', '{}']
sh: expand=True
  echo {} {{i}}
""".format(
                    tag1, tag2, tag1
                )
            )
        wf = SoS_Script(filename="test_tags.sos").workflow()
        Base_Executor(
            wf,
            config={
                "sig_mode": "force",
                "script": "test_trunkworker.sos",
                "max_running_jobs": 10,
                "workflow_args": [],
                "output_dag": "",
                "output_report": None,
                "targets": [],
                "worker_procs": ["4"],
                "default_queue": "localhost",
                "workflow": "default",
                "workdir": ".",
            },
        ).run()
        ret = subprocess.check_output(
            "sos status -t {}".format(tag2), shell=True
        ).decode()
        self.assertEqual(len(ret.splitlines()), 2, "Obtained {}".format(ret))

    @unittest.skipIf(not has_docker, "Docker container not usable")
    def test_max_mem(self):
        """Test server restriction max_mem"""
        script = SoS_Script(
            """
[10]
task: mem='2G'
print('a')
"""
        )
        wf = script.workflow()
        self.assertRaises(
            Exception,
            Base_Executor(
                wf,
                config={
                    "config_file": "~/docker.yml",
                    "default_queue": "docker_limited",
                    "sig_mode": "force",
                },
            ).run,
        )

    def test_local_runtime_max_walltime(self):
        """Test server max_walltime option"""
        script = SoS_Script(
            """
[10]
task:
import time
time.sleep(15)
"""
        )
        wf = script.workflow()
        self.assertRaises(
            Exception,
            Base_Executor(
                wf,
                config={
                    "config_file": "~/docker.yml",
                    "default_queue": "local_limited",
                    "sig_mode": "force",
                },
            ).run,
        )

    @unittest.skipIf(not has_docker, "Docker container not usable")
    def test_max_cores(self):
        """Test server restriction max_cores"""
        script = SoS_Script(
            """
[10]
task: cores=8
print('a')
"""
        )
        wf = script.workflow()
        self.assertRaises(
            Exception,
            Base_Executor(
                wf,
                config={
                    "config_file": "~/docker.yml",
                    "default_queue": "docker_limited",
                    "sig_mode": "force",
                },
            ).run,
        )

    @unittest.skipIf(not has_docker, "Docker container not usable")
    def test_override_max_cores(self):
        """Test use queue_args to override server restriction max_cores"""
        script = SoS_Script(
            """
[10]
task: cores=8
print('a')
"""
        )
        wf = script.workflow()
        Base_Executor(
            wf,
            config={
                "config_file": "~/docker.yml",
                "default_queue": "docker_limited",
                "sig_mode": "force",
                "queue_args": {"cores": 1},
            },
        ).run()

    def test_list_hosts(self):
        """test list hosts using sos status -q"""
        for v in ["0", "1", "3", "4"]:
            output = subprocess.check_output(
                ["sos", "remote", "list", "-c", "~/docker.yml", "-v", v]
            ).decode()
            self.assertTrue(
                "local_limited" in output,
                f"local_limited not in \n{output}\n for verbosity {v}",
            )

    @unittest.skipIf(not has_docker, "Docker container not usable")
    def test_max_walltime(self):
        """Test server restriction max_walltime"""
        script = SoS_Script(
            """
[10]
task: walltime='1:00:00'
print('a')
"""
        )
        wf = script.workflow()
        self.assertRaises(
            Exception,
            Base_Executor(
                wf,
                config={
                    "config_file": "~/docker.yml",
                    "default_queue": "docker_limited",
                    "sig_mode": "force",
                },
            ).run,
        )

    def test_purge_all_with_option(self):
        """Test sos purge all with options such as -s completed"""
        with cd_new("temp_c"):
            with open("test.sos", "w") as tst:
                tst.write(
                    """
input: for_each={'i': range(2)}
output: f'a{i}.txt'
task:
run: expand=True
echo temp_a
touch {_output}
"""
                )
            subprocess.call("sos run test -s force -q localhost", shell=True)
            tasks = get_tasks()
            subprocess.call("sos purge --all -s failed", shell=True)
        # check tasks
        taskstatus = [
            x.split()[0]
            for x in subprocess.check_output("sos status -v1", shell=True)
            .decode()
            .splitlines()
        ]
        self.assertTrue(all(x in taskstatus for x in tasks))
        # purge one of them
        subprocess.call(f"sos purge {tasks[0]}", shell=True)
        taskstatus = [
            x.split()[0]
            for x in subprocess.check_output("sos status -v1", shell=True)
            .decode()
            .splitlines()
        ]
        self.assertTrue(tasks[0] not in taskstatus)
        self.assertTrue(tasks[1] in taskstatus)
        #
        subprocess.call("sos purge --all", shell=True)
        taskstatus = [
            x.split()[0]
            for x in subprocess.check_output("sos status -v1", shell=True)
            .decode()
            .splitlines()
        ]
        self.assertTrue(tasks[1] not in taskstatus)

    def test_resubmit_task_with_different_walltime(self):
        """Test resubmission of tasks with different walltime #1019"""
        with cd_new("temp_walltime"):
            with open("test.sos", "w") as tst:
                tst.write(
                    """
task: walltime='1m'
sh:
echo 0.1
"""
                )
            subprocess.call("sos run test -s force -q localhost", shell=True)
            tasks = get_tasks()
            out = subprocess.check_output(f"sos status {tasks[0]} -v4", shell=True)
            self.assertTrue("00:01:00" in out.decode())
            with open("test1.sos", "w") as tst:
                tst.write(
                    """
task: walltime='2m'
sh:
echo 0.1
"""
                )
            subprocess.call("sos run test1 -s force -q localhost", shell=True)
            new_tasks = get_tasks()
            self.assertEqual(tasks, new_tasks)
            #
            out = subprocess.check_output(f"sos status {tasks[0]} -v4", shell=True)
            self.assertTrue("00:02:00" in out.decode())

    def test_task_signature(self):
        """Test re-execution of tasks"""
        with cd_new("temp_signature"):
            with open("test.sos", "w") as tst:
                tst.write(
                    """
task:
sh:
sleep 2
"""
                )
            subprocess.call("sos run test -s force -q localhost", shell=True)
            tasks = get_tasks()
            tf = TaskFile(tasks[0])
            self.assertTrue(tf.has_signature())
            self.assertEqual(tf.status, "completed")
            tf.tags_created_start_and_duration()
            #
            subprocess.call("sos run test -q localhost", shell=True)
            self.assertLess(tf.tags_created_start_and_duration()[3], 1)

    def test_wrong_host(self):
        script = SoS_Script(
            """
[10]
task: walltime='1:00:00', queue='undefined'
print('a')
"""
        )
        wf = script.workflow()
        self.assertRaises(
            Exception, Base_Executor(wf, config={"default_queue": "localhost"}).run
        )

    def test_output_in_task(self):
        """Test passing _output to task #1136"""
        script = SoS_Script(
            """
chunks  = [1,2]
[1]
input: for_each = 'chunks'
output: f'{_chunks}.txt'
_output.touch()

[2]
input: group_with = 'chunks'
output: summary_stats = f'{_input}.summary', ld_matrix = f'{_input}.result'
task:

python3: expand="${ }"
       open("${_output['summary_stats']}", 'w').close()
       open("${_output['ld_matrix']}", 'w').close()
"""
        )
        wf = script.workflow()
        Base_Executor(wf, config={"default_queue": "localhost"}).run()

    def test_repeated_tasks(self):
        """Test statement before task #1142 """
        script = SoS_Script(
            """
[1]
input: for_each=dict(i=range(5))

print(f'This is for {i}')
task:  walltime='10m'
print(f'this is task {i}')

"""
        )
        for i in range(5):
            wf = script.workflow()
            Base_Executor(
                wf, config={"sig_mode": "force", "default_queue": "localhost"}
            ).run()

    def test_passing_parameters_to_task(self):
        """Test passing of parameters in global section to tasks #1155"""
        script = SoS_Script(
            """\
[global]
parameter: par=5
def a():
  print(par)

[default]
task:
a()
"""
        )
        wf = script.workflow()
        Base_Executor(
            wf, config={"sig_mode": "force", "default_queue": "localhost"}
        ).run()

    def test_trunk_size_with_stop_if(self):
        """Test a case when some tasks are not submitted due to holes in slots #1159"""
        for i in range(5):
            f = f"{i+1}.txt"
            if os.path.isfile(f):
                os.remove(f)
        script = SoS_Script(
            """\
[1]
output: [f'{x+1}.txt' for x in range(5)]
for i in range(5):
  name = f'{i+1}.txt'
  if i not in [0,1,2]:
    path(name).touch()
  else:
    with open(name, 'w') as f:
      f.write('test it')

[2]
input: group_by = 1
output: f'{_input:n}.out'
stop_if(_input.stat().st_size==0, no_output=True)

task: trunk_size = 80
_output.touch()
"""
        )
        wf = script.workflow()
        Base_Executor(
            wf, config={"sig_mode": "force", "default_queue": "localhost"}
        ).run()

    def test_output_from_master_task(self):
        """Test splitting the output from master task #1203"""
        script = SoS_Script(
            """\
l=[x for x in range(1,13)]

[2]
input: for_each = 'l'
output: f'{_l}.out'

task: trunk_size = 4
_output.touch()

[3]
assert _input == f'{_index+1}.out'
"""
        )
        wf = script.workflow()
        Base_Executor(wf, config={"default_queue": "localhost"}).run()

    @unittest.skipIf(not has_docker, "Docker container not usable")
    def test_sync_input_output_and_rerun(self):
        """Test sync input and output with remote host"""
        for i in range(4):
            if os.path.isfile(f"test_{i}.txt"):
                os.remove(f"test_{i}.txt")
            if os.path.isfile(f"test_{i}.bak"):
                os.remove(f"test_{i}.bak")
        import random

        script = SoS_Script(
            """
parameter: g = 100

[10]
input: for_each=dict(i=range(4))
output: f'test_{i}.txt'

with open(f'test_{i}.txt', 'w') as tst:
    tst.write(f'test_{i}_{g}')

[20]
output: _input.with_suffix('.bak')

task:

with open(_input, 'r') as inf, open(_output, 'w') as outf:
	outf.write(inf.read() + '.bak')
"""
        )
        wf = script.workflow()
        val = random.randint(1, 10000)
        Base_Executor(
            wf,
            args=["--g", str(val)],
            config={
                "config_file": "~/docker.yml",
                "default_queue": "docker",
                "sig_mode": "force",
            },
        ).run()
        # now check if
        for i in range(4):
            self.assertTrue(os.path.isfile(f"test_{i}.txt"))
            with open(f"test_{i}.bak") as outf:
                self.assertEqual(outf.read(), f"test_{i}_{val}.bak")
            self.assertTrue(os.path.isfile(f"test_{i}.bak"))
            with open(f"test_{i}.bak") as outf:
                self.assertEqual(outf.read(), f"test_{i}_{val}.bak")
        #
        # test rerun the task file on local host
        for i in range(4):
            if os.path.isfile(f"test_{i}.txt"):
                os.remove(f"test_{i}.txt")
            if os.path.isfile(f"test_{i}.bak"):
                os.remove(f"test_{i}.bak")
        Base_Executor(wf, args=["--g", str(val)], config={"sig_mode": "force"}).run()
        for i in range(4):
            self.assertTrue(os.path.isfile(f"test_{i}.txt"))
            with open(f"test_{i}.bak") as outf:
                self.assertEqual(outf.read(), f"test_{i}_{val}.bak")
            self.assertTrue(os.path.isfile(f"test_{i}.bak"))
            with open(f"test_{i}.bak") as outf:
                self.assertEqual(outf.read(), f"test_{i}_{val}.bak")

    @unittest.skipIf(not has_docker, "Docker container not usable")
    def test_remote_input_target(self):
        """Test the use of remote target"""
        if os.path.isfile("vars.sh"):
            os.remove("vars.sh")
        if os.path.isfile("vars1.sh"):
            os.remove("vars1.sh")
        script = SoS_Script(
            """

[10]
input: remote('/lib/init/vars.sh')
output: f'vars1.sh'

task:

with open(_input, 'r') as inf, open(_output, 'w') as outf:
	outf.write(inf.read())
"""
        )
        wf = script.workflow()
        Base_Executor(
            wf,
            config={
                "config_file": "~/docker.yml",
                "default_queue": "docker",
                "sig_mode": "force",
            },
        ).run()
        self.assertFalse(os.path.isfile("vars.sh"))
        self.assertTrue(os.path.isfile("vars1.sh"))

    @unittest.skipIf(not has_docker, "Docker container not usable")
    def test_delayed_interpolation(self):
        """Test delayed interpolation with expression involving remote objects"""
        # purge all previous tasks
        if file_target("test.py").exists():
            file_target("test.py").unlink()
        if file_target("test.py.bak").exists():
            file_target("test.py.bak").unlink()
        script = SoS_Script(
            """
[10]
output: remote('test.py')
task:
run:
    touch test.py

[20]
output: remote(f"{_input:R}.bak")
task:
run: expand=True
    cp {_input} {_output}
"""
        )
        wf = script.workflow()
        Base_Executor(
            wf,
            config={
                "config_file": "~/docker.yml",
                # do not wait for jobs
                "wait_for_task": True,
                "default_queue": "docker",
                "sig_mode": "force",
            },
        ).run()
        # this file is remote only
        self.assertFalse(os.path.isfile("test.py"))
        self.assertFalse(os.path.isfile("test.py.bak"))


@pytest.mark.skipif(not has_docker, reason="Docker container not usable")
def test_remote_output_target(clear_now_and_after):
    """Test the use of remote target"""
    clear_now_and_after("vars.sh", "vars1.sh")
    execute_workflow(
        """
        [10]
        input: remote('/lib/init/vars.sh')
        output: remote(f'vars1.sh')

        task:

        with open(_input, 'r') as inf, open(_output, 'w') as outf:
            outf.write(inf.read())
        """,
        options={
            "config_file": "~/docker.yml",
            "default_queue": "docker",
            "sig_mode": "force",
        },
    )
    assert not os.path.isfile("vars.sh")
    assert not os.path.isfile("vars1.sh")


@pytest.mark.skipif(not has_docker, reason="Docker container not usable")
def test_remote_output_target_with_trunksize(clear_now_and_after):
    clear_now_and_after("vars.sh", "vars1.sh")

    execute_workflow(
        """\
        [10]
        import os
        input: remote('/lib/init/vars.sh'), remote('/lib/init/init-d-script'), group_by=1
        output: remote(os.path.basename(str(_input)))

        task: trunk_size=2

        with open(_input, 'r') as inf, open(_output, 'w') as outf:
            outf.write(inf.read())""",
        options={
            "config_file": "~/docker.yml",
            "default_queue": "docker",
            "sig_mode": "force",
        },
    )
    assert not os.path.isfile("vars.sh")
    assert not os.path.isfile("init-d-script")


def test_runtime_max_walltime():
    """Test server max_walltime option"""
    with pytest.raises(Exception):
        execute_workflow(
            """
        [10]
        task:
        import time
        time.sleep(25)
        """,
            options={
                "config_file": "~/docker.yml",
                "default_queue": "docker_limited",
                "sig_mode": "force",
            },
        )


@pytest.mark.skipif(not has_docker, reason="Docker container not usable")
def test_sync_master_task(clear_now_and_after):
    """Test sync input and output with remote host with trunksize"""
    clear_now_and_after(
        [f"test_{i}.txt" for i in range(4)], [f"test_{i}.bak" for i in range(4)]
    )
    import random

    val = random.randint(1, 10000)
    execute_workflow(
        r"""
        parameter: g = 100

        [10]
        input: for_each=dict(i=range(4))
        output: f'test_{i}.txt'

        with open(f'test_{i}.txt', 'w') as tst:
            tst.write(f'test_{i}_{g}')

        [20]
        output: _input.with_suffix('.bak')

        task: trunk_size=2

        with open(_input, 'r') as inf, open(_output, 'w') as outf:
            outf.write(inf.read() + '.bak')
        """,
        args=["--g", str(val)],
        options={
            "config_file": "~/docker.yml",
            "default_queue": "docker",
            "sig_mode": "force",
        },
    )
    # now check if
    for i in range(4):
        assert os.path.isfile(f"test_{i}.txt")
        with open(f"test_{i}.bak") as outf:
            assert outf.read() == f"test_{i}_{val}.bak"
        assert os.path.isfile(f"test_{i}.bak")
        with open(f"test_{i}.bak") as outf:
            assert outf.read() == f"test_{i}_{val}.bak"


def test_local_max_cores():
    """Test server restriction max_cores"""
    with pytest.raises(Exception):
        execute_workflow(
            """
        [10]
        task: cores=8
        print('a')
        """,
            options={
                "config_file": "~/docker.yml",
                "default_queue": "local_limited",
                "sig_mode": "force",
            },
        )


def test_local_max_mem():
    """Test server restriction max_mem"""
    with pytest.raises(Exception):
        execute_workflow(
            """
        [10]
        task: mem='2G'
        print('a')
        """,
            options={
                "config_file": "~/docker.yml",
                "default_queue": "local_limited",
                "sig_mode": "force",
            },
        )



#!/usr/bin/env python
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import os
import signal
import sys
import threading
from contextlib import contextmanager

import pytest

from sos import execute_workflow

try:
    import _thread
except Exception:
    import _dummy_thread as _thread

try:
    from sos.docker.client import SoS_DockerClient
    has_docker = True
except ImportError:
    print('Docker is not available')
    has_docker = False


class TimeoutException(Exception):

    def __init__(self, msg=''):
        self.msg = msg


@contextmanager
def time_limit(seconds, msg=''):
    if sys.platform == 'win32':
        # windows system does not have signal SIGALARM so we will
        # have to use a timer approach, which does not work as well
        # as the signal approach
        def timeout_func():
            #env.logger.error('Timed out for operation {}'.format(msg))
            _thread.interrupt_main()

        timer = threading.Timer(seconds, timeout_func)
        timer.start()
        try:
            yield
        except KeyboardInterrupt:
            # important: KeyboardInterrupt does not interrupt time.sleep()
            # because KeyboardInterrupt is handled by Python interpreter but
            # time.sleep() calls a system function.
            raise TimeoutException("Timed out for operation {}".format(msg))
        finally:
            # if the action ends in specified time, timer is canceled
            timer.cancel()
    else:

        def signal_handler(signum, frame):
            raise TimeoutException("Timed out for option {}".format(msg))

        signal.signal(signal.SIGALRM, signal_handler)
        signal.alarm(seconds)
        try:
            yield
        finally:
            signal.alarm(0)


try:
    with time_limit(2, 'check docker daemon'):
        has_docker = SoS_DockerClient().client is not None
except Exception as e:
    print(
        'Cannot connect to a docker daemon in 2 seconds. Assuming no docker environment.'
    )
    print(e)
    has_docker = False


@pytest.mark.skipif(not has_docker or sys.platform == 'win32',
                    reason='Skip test because docker is not installed.')
def test_bash_in_docker():
    '''Test action bash in docker environment'''
    execute_workflow(r'''
        [0]
        run:  container='docker://ubuntu'
        echo 'Echo'
        ''')


#    @pytest.mark.skipif(not has_docker or sys.platform != 'win32', reason='Skip test because docker is not installed.')
#    def testBatchScriptInDocker():
#        '''Test action powershell in docker environment'''
#        script = execute_workflow(r'''
#           [0]
#           run:  container='docker://microsoft/windowsservercore'
#           dir c:\
#           ''')

@pytest.mark.skipif(not has_docker or sys.platform == 'win32',
                    reason='Skip test because docker is not installed.')
def test_sh_in_docker():
    '''Test action sh in docker environment'''
    # test docker
    script = execute_workflow(r'''
        [0]
        run: container='docker://ubuntu'
        echo 'Echo
        ''',
        options={'run_mode': 'dryrun'})
    with pytest.raises(Exception):
        execute_workflow(script)

@pytest.mark.skipif(not has_docker or sys.platform == 'win32',
                    reason='Skip test because docker is not installed.')
def test_docker_build_linux_image():
    '''Test action docker build'''
    execute_workflow(r'''
        [0]
        docker_build:  tag='test/docker_build'
        #
        # Super simple example of a Dockerfile
        #
        FROM ubuntu:latest
        MAINTAINER Andrew Odewahn "odewahn@oreilly.com"

        RUN apt-get update

        WORKDIR /home
        ''')


@pytest.mark.skipif(not has_docker or sys.platform == 'win32',
                 reason='Skip test because docker is not installed.')
def test_docker_build_linux_image_option_label_compress():
    '''Test action docker build'''
    # build with more options
    execute_workflow(r'''
        [0]
        docker_build:  tag='test/docker_build1', label='my_label', compress=True, memory='2G'
        #
        # Super simple example of a Dockerfile
        #
        FROM ubuntu:latest
        MAINTAINER Andrew Odewahn "odewahn@oreilly.com"

        WORKDIR /home
        ''')


@pytest.mark.skipif(not has_docker or sys.platform == 'win32' or
                 'TRAVIS' in os.environ,
                 reason='Skip test because docker is not installed.')
def test_docker_build_linux_image_label_with_space():
    '''Test action docker build'''
    with pytest.raises(Exception):
        execute_workflow(r'''
            [0]
            docker_build:  tag='test/docker_build1', label='my label with space', compress=True, memory='2G'
            #
            # Super simple example of a Dockerfile
            #
            FROM ubuntu:latest
            MAINTAINER Andrew Odewahn "odewahn@oreilly.com"

            WORKDIR /home
            ''')


@pytest.mark.skipif(not has_docker or sys.platform != 'win32' or
                 'APPVEYOR' in os.environ,
                 reason='Skip test because docker is not installed.')
def test_docker_build_windows_image():
    '''Test action docker build'''
    execute_workflow(r'''
        [0]
        docker_build:  tag='test/docker_build'
        # Indicates that the windowsservercore image will be used as the base image.
        FROM microsoft/windowsservercore

        # Metadata indicating an image maintainer.
        MAINTAINER someone@microsoft.com

        ''')


@pytest.mark.skipif(not has_docker or sys.platform == 'win32',
                    reason='Skip test because docker is not installed.')
def test_docker_image(fastq_files):
    '''Test docker_image option'''
    execute_workflow(r'''
        import os
        import glob
        [0]
        fastq_files = glob.glob('data/S20_R*.fastq')
        input_volume = os.path.dirname(os.path.abspath(fastq_files[0]))
        output_volume = os.getcwd()

        run: container='docker://compbio/ngseasy-fastqc:1.0-r001',
            volumes=[f"{input_volume}:/input_data", f"{output_volume}:/output_data"]

            ls -l /input_data
            /usr/local/bin/fastqc /input_data/*.fastq --outdir /output_data
        ''')


@pytest.mark.skipif(
    not has_docker or sys.platform == 'win32',
    reason='Skip test because docker is not installed, or in travis, which failed for unknown reason'
)
def test_docker_image_from_file():
    '''Test docker_image load from a file.'''
    # image from a saved file
    execute_workflow(r'''
        [0]
        run:   container='docker://blang/busybox-bash'

        [1]
        run:
            docker save blang/busybox-bash > hello.tar
            docker rmi -f blang/busybox-bash

        [2]
        run: container='docker://blang/busybox-bash', docker_file = 'hello.tar'

            echo "a"
        ''')


@pytest.mark.skipif(not has_docker or sys.platform == 'win32',
                    reason='Skip test because docker is not installed.')
def test_docker_script_action():
    '''Test action sh in docker environment'''
    # test docker
    execute_workflow(r'''
        [0]
        script: container='docker://ubuntu', args='{script}'
        echo 'Echo'
        ''')

@pytest.mark.skipif(not has_docker or sys.platform == 'win32',
                    reason='Skip test because docker is not installed.')
def test_port_option():
    '''Test use of option port in action'''
    execute_workflow(r'''
        [0]
        run:  container='ubuntu', port=True
        echo 'Echo'
        ''')
    #
    execute_workflow(r'''
        [0]
        run:  container='ubuntu', port=2345
        echo 'Echo'
        ''')

@pytest.mark.skipif(not has_docker or sys.platform == 'win32',
                    reason='Skip test because docker is not installed.')
def test_auto_output_mount():
    '''Test use of option port in action'''
    execute_workflow(r'''
        output: '../data/1.txt'
        run: container='ubuntu'
        sh: expand=True
            touch {_output}
        ''')
    assert os.path.isfile('../data/1.txt')


#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import os
import pytest
import subprocess
import unittest

from sos.parser import ParsingError, SoS_Script
from sos.targets import file_target, sos_targets, path, paths
from sos.utils import ArgumentError, env
from sos.workflow_executor import Base_Executor
from sos.converter import extract_workflow
from sos import execute_workflow

section1_sos = """
#!/usr/bin/env sos-runner
#fileformat=SOS1.1

#
# this is a test sos script
#
var1='value1'
var2 = 'value2'
var3 = [var1,
  var2]

# section
# description for workflow section
#
[parameters]
# par1 string
par1 = 'var1'

# par2 list
par2 = ['a', 'b', 'c']

# par3 multiline
par3 = ['a',
	'b']

[*_0]
var0 = '0'

[section_10]
#
#step 10
var1 = 'a'

[section_2 : shared='var3']
var2 = 'a'
input: var1
output: var2

var3 = 'a'

[section_3, *_4 : shared='var4']
output:
	var2,
	var3

print()
var4= 'value4'

[chapter_5]
var4='5'
 """


class TestParser(unittest.TestCase):
    def setUp(self):
        env.reset()
        subprocess.call("sos remove -s", shell=True)
        self.temp_files = []

    def tearDown(self):
        for f in self.temp_files:
            if file_target(f).exists():
                file_target(f).unlink()

    def touch(self, files):
        """create temporary files"""
        if isinstance(files, str):
            files = [files]
        #
        for f in files:
            with open(f, "w") as tmp:
                tmp.write("test")
        #
        self.temp_files.extend(files)

    def test_file_format(self):
        """Test recognizing the format of SoS script"""
        # file format must be 'fileformat=SOSx.x'
        self.assertRaises(ParsingError, SoS_Script, "#fileformat=SS2")
        self.assertRaises(ParsingError, SoS_Script, "#fileformat=SOS1.0beta")
        #
        # parse a larger script with gormat 1.1
        SoS_Script(section1_sos)
        # not the default value of 1.0
        # self.assertEqual(script.format_version, '1.1')

    def test_workflows(self):
        """Test workflows defined in SoS script"""
        script = SoS_Script("""[0]""")
        self.assertEqual(sorted(script.workflows), [""])
        script = SoS_Script("""[0]\n[1]""")
        self.assertEqual(sorted(script.workflows), [""])
        script = SoS_Script("""[0]\n[*_1]""")
        self.assertEqual(sorted(script.workflows), [""])
        script = SoS_Script("""[0]\n[*_1]\n[auxiliary:provides='{a}.txt']""")
        self.assertEqual(sorted(script.workflows), ["", "auxiliary"])
        script = SoS_Script("""[0]\n[*_1]\n[human_1]""")
        self.assertEqual(sorted(script.workflows), ["", "human"])
        script = SoS_Script("""[0]\n[*_1]\n[human_1]\n[mouse_2]""")
        self.assertEqual(sorted(script.workflows), ["", "human", "mouse"])
        script = SoS_Script("""[0]\n[*_1]\n[human_1]\n[mouse_2]\n[s*_2]""")
        self.assertEqual(sorted(script.workflows), ["", "human", "mouse"])
        # unnamed
        script = SoS_Script("""[0]\n[*_1]\n[human_1]\n[mouse]\n[s*_2]""")
        self.assertEqual(sorted(script.workflows), ["", "human", "mouse"])
        #
        # workflow name with -
        script = SoS_Script("""[proc-1]\n[test-case_2]""")
        self.assertEqual(sorted(script.workflows), ["proc-1", "test-case"])
        script.workflow("proc-1")
        script.workflow("proc-1 + test-case:2")

    def test_sections(self):
        """Test section definitions"""
        # bad names
        for badname in ["56_1", "_a", "a_", "1x", "*", "?"]:
            self.assertRaises(ParsingError, SoS_Script, "[{}]".format(badname))
        # bad options
        for badoption in ["ss"]:
            self.assertRaises(ParsingError, SoS_Script, "[0:{}]".format(badoption))
        # allowed names
        for name in ["a5", "a_5", "*_0", "a*1_100"]:
            SoS_Script("[{}]".format(name))
        # allowed names with alias
        for name in ["a5 (p1)", "a_5 (something fun)", "*_0 (no way)", "a*1_100"]:
            SoS_Script("[{}]".format(name))
        # duplicate sections
        self.assertRaises(ParsingError, SoS_Script, """[1]\n[1]""")
        self.assertRaises(ParsingError, SoS_Script, """[1]\n[3]\n[2,1]""")
        self.assertRaises(ParsingError, SoS_Script, """[a_1]\n[a_3]\n[*_1]""")
        #
        # no duplicated section header
        SoS_Script("""[a_1]\n[a_3]\n[b*_1]""")
        #
        # global section
        self.assertRaises(ParsingError, SoS_Script, """[global, step_10]""")

    def test_parameters(self):
        """Test ending parameters with new line #1311"""
        self.assertRaises(ParsingError, SoS_Script, """input: ['a.txt'\n\n'b.txt']\n""")

    def test_global_variables(self):
        """Test definition of variables"""
        # allow definition
        SoS_Script("""a = '1' """)
        SoS_Script("""a = ['a', 'b'] """)
        # but this one has incorrect syntax
        self.assertRaises(ParsingError, SoS_Script, """a = 'b  """)
        # This one also does not work because b is not defined.
        # delattr(env, 'sos_dict')
        # script = SoS_Script('''a = b\n[0] ''')
        # wf = script.workflow()
        # dag = Base_Executor(wf).run(mode='dryrun')
        # self.assertRaises(ValueError, Base_Executor(wf).run, dag)
        # multi-line string literal
        SoS_Script(
            '''a = """
this is a multi line
string """
'''
        )
        # multi-line list literal, even with newline in between
        SoS_Script(
            """a = [
'a',

'b'
]
"""
        )
        #
        SoS_Script(section1_sos)
        # not the default value of 1.0
        #
        script = SoS_Script(
            """
[global]
a = 1

[b]
print(a)
"""
        )
        wf = script.workflow()
        Base_Executor(wf).run()

    def test_parameters_section(self):
        """Test parameters section"""
        # directive not allowed in parameters
        script = SoS_Script(section1_sos)
        wf = script.workflow("chapter:0")
        # self.assertRaises(ArgumentError, Base_Executor(wf).run,
        #    args=['--not_exist'])
        # self.assertRaises(ArgumentError, Base_Executor(wf).run,
        #    args=['--par1', 'a', 'b'])
        #
        script = SoS_Script(
            """
parameter: a = [1, 2]
[0]
"""
        )
        wf = script.workflow()
        self.assertEqual(list(wf.parameters().keys()), ["a"])
        Base_Executor(wf).run()
        self.assertEqual(env.sos_dict["a"], [1, 2])
        env.sos_dict.pop("a")
        wf = script.workflow()
        Base_Executor(wf, args=["--a", "3"]).run()
        self.assertEqual(env.sos_dict["a"], [3])
        env.sos_dict.pop("a")
        wf = script.workflow()
        Base_Executor(wf, args=["--a", "3", "5"]).run()
        self.assertEqual(env.sos_dict["a"], [3, 5])
        env.sos_dict.pop("a")
        #
        script = SoS_Script(
            """
# comment
# comment
parameter: a = ['a.txt', 'b.txt']
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf).run()
        self.assertEqual(env.sos_dict["a"], ["a.txt", "b.txt"])
        env.sos_dict.pop("a")
        wf = script.workflow()
        Base_Executor(wf, args=["--a", "3"]).run()
        self.assertEqual(env.sos_dict["a"], ["3"])
        env.sos_dict.pop("a")
        wf = script.workflow()
        Base_Executor(wf, args=["--a", "3", "5"]).run()
        self.assertEqual(env.sos_dict["a"], ["3", "5"])
        env.sos_dict.pop("a")
        #
        # test parameter using global definition
        script = SoS_Script(
            """
a="100"

# comment
parameter: b=str(int(a)+1)
[0]
"""
        )
        wf = script.workflow()
        self.assertEqual(list(wf.parameters().keys()), ["b"])
        Base_Executor(wf).run()
        self.assertEqual(env.sos_dict["b"], "101")
        env.sos_dict.pop("b")
        #
        env.sos_dict.clear()
        script = SoS_Script(
            """
a=100

parameter: b=a+1
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf).run()
        self.assertEqual(env.sos_dict["b"], 101)
        env.sos_dict.pop("b")
        #
        script = SoS_Script(
            """
a=100

parameter: b=a+1.
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--b", "1000"]).run()
        #
        self.assertEqual(env.sos_dict["b"], 1000)
        env.sos_dict.pop("b")
        #
        # argument has hve a value
        self.assertRaises(
            ParsingError,
            SoS_Script,
            """
[0]
parameter: b=

""",
        )
        # if it is a type, must provide value
        script = SoS_Script(
            """
# comment
parameter: b = int
[0]
"""
        )
        wf = script.workflow()
        self.assertRaises(ArgumentError, Base_Executor, wf)
        #
        script = SoS_Script(
            """
parameter: b = list
[0]
"""
        )
        self.assertEqual(list(wf.parameters().keys()), ["b"])
        wf = script.workflow()
        self.assertRaises(ArgumentError, Base_Executor, wf)
        # also require the type
        script = SoS_Script(
            """
parameter: b = int
[0]
"""
        )
        wf = script.workflow()
        self.assertRaises(ArgumentError, Base_Executor, wf, args=["--b", "file"])
        #
        script = SoS_Script(
            """
parameter: b = int
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--b", "5"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["b"], 5)
        env.sos_dict.pop("b")
        # string
        script = SoS_Script(
            """
parameter: b = str
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--b", "5"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["b"], "5")
        env.sos_dict.pop("b")
        # list is ok
        script = SoS_Script(
            """
parameter: b = list
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--b", "5"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["b"], ["5"])
        # bool required
        script = SoS_Script(
            """
# comment
parameter: b = bool
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--b"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["b"], True)
        env.sos_dict.pop("b")
        Base_Executor(wf, args=["--no-b"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["b"], False)
        env.sos_dict.pop("b")
        # bool with default True
        script = SoS_Script(
            """
parameter: b = True
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=[]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["b"], True)
        env.sos_dict.pop("b")
        Base_Executor(wf, args=["--b"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["b"], True)
        env.sos_dict.pop("b")
        Base_Executor(wf, args=["--no-b"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["b"], False)
        env.sos_dict.pop("b")
        # bool with default False
        script = SoS_Script(
            """
parameter: b = False
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=[]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["b"], False)
        env.sos_dict.pop("b")
        Base_Executor(wf, args=["--b"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["b"], True)
        env.sos_dict.pop("b")
        Base_Executor(wf, args=["--no-b"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["b"], False)
        env.sos_dict.pop("b")
        #
        # parameters cannot coincide with a readonly global variable
        # are masked by previous definition
        script = SoS_Script(
            """
a = 4
parameter: a = 5
[0]
"""
        )
        wf = script.workflow()
        self.assertRaises(Exception, Base_Executor, wf, args=["--a", 7])
        # self.assertEqual(env.sos_dict['a'], 4)
        #
        # test parameters with dash
        script = SoS_Script(
            """
parameter: a_b = 5
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--a_b", "10"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["a_b"], 10)
        env.sos_dict.pop("a_b")
        Base_Executor(wf, args=["--a-b", "10"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["a_b"], 10)
        env.sos_dict.pop("a_b")
        #
        #
        script = SoS_Script(
            """
parameter: a_b = int
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--a_b", "10"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["a_b"], 10)
        env.sos_dict.pop("a_b")
        Base_Executor(wf, args=["--a-b", "10"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["a_b"], 10)
        env.sos_dict.pop("a_b")
        #
        # test support for type path, paths, file_target and sos_targets
        script = SoS_Script(
            """
parameter: path_var = path
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--path-var", "a.txt"]).run(mode="run")
        self.assertTrue(isinstance(env.sos_dict["path_var"], path))
        #
        script = SoS_Script(
            """
parameter: path_var = file_target
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--path-var", "a.txt"]).run(mode="run")
        self.assertTrue(isinstance(env.sos_dict["path_var"], file_target))
        #
        script = SoS_Script(
            """
parameter: path_var = paths
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--path-var", "a.txt", "b.txt"]).run(mode="run")
        self.assertTrue(isinstance(env.sos_dict["path_var"], paths))
        #
        #
        script = SoS_Script(
            """
parameter: path_var = sos_targets
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--path-var", "a.txt"]).run(mode="run")
        self.assertTrue(isinstance(env.sos_dict["path_var"], sos_targets))

        #
        script = SoS_Script(
            """
parameter: path_var = path('a.txt')
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--path-var", "a.txt"]).run(mode="run")
        self.assertTrue(isinstance(env.sos_dict["path_var"], path))
        #
        script = SoS_Script(
            """
parameter: path_var = file_target('a.txt')
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--path-var", "a.txt"]).run(mode="run")
        self.assertTrue(isinstance(env.sos_dict["path_var"], file_target))
        #
        script = SoS_Script(
            """
parameter: path_var = paths('a.txt')
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--path-var", "a.txt", "b.txt"]).run(mode="run")
        self.assertTrue(isinstance(env.sos_dict["path_var"], paths))
        #
        #
        script = SoS_Script(
            """
parameter: path_var = sos_targets('a.txt')
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--path-var", "a.txt"]).run(mode="run")
        self.assertTrue(isinstance(env.sos_dict["path_var"], sos_targets))

        #
        # Test allow the use of sos keywords as parameters #1041
        script = SoS_Script(
            """\
[1]
parameter: input = 5
output = 10
python: expand=True
  print({input})
"""
        )
        wf = script.workflow()
        Base_Executor(wf).run()
        # multiple parameters
        script = SoS_Script(
            """
parameter: a_b = int
[0]
parameter: c_b = list
"""
        )
        wf = script.workflow()
        self.assertEqual(sorted(list(wf.parameters().keys())), ["a_b", "c_b"])

    def test_param_in_task(self):
        """Test specification of parameters in tasks"""
        self.assertRaises(
            Exception,
            SoS_Script,
            """\
[1]
output: 'a.txt'
task:
parameter: value = 'a'
sh: expand=True
echo {value} >  a.txt
""",
        )
        # multiple parameters
        script = SoS_Script(
            """
parameter: a_b = int
[0]
parameter: c_b = list
"""
        )
        wf = script.workflow()
        self.assertEqual(sorted(list(wf.parameters().keys())), ["a_b", "c_b"])

    def test_type_trait_parameter(self):
        # type trait
        script = SoS_Script(
            """
parameter: b
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--b", "5"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["b"], "5")
        #
        script = SoS_Script(
            """
parameter: b :str
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--b", "5"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["b"], "5")
        #
        script = SoS_Script(
            """
parameter: b : list
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--b", "5"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["b"], ["5"])

        #
        script = SoS_Script(
            """
parameter: b : list = 5
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--b", "5"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["b"], 5)
        #
        script = SoS_Script(
            """
parameter: b : int
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--b", "5"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["b"], 5)

    def test_input_target(self):
        # test input of targets
        script = SoS_Script(
            """
parameter: b : file_target
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--b", "aaa"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["b"].__class__.__name__, "file_target")
        #
        script = SoS_Script(
            """
parameter: b = file_target('file')
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--b", "aaa"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["b"].__class__.__name__, "file_target")
        #
        script = SoS_Script(
            """
parameter: b : sos_targets
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--b", "aaa", "bbb"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["b"].__class__.__name__, "sos_targets")
        #
        script = SoS_Script(
            """
parameter: b = sos_targets('file')
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--b", "aaa"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["b"].__class__.__name__, "sos_targets")
        #
        script = SoS_Script(
            """
parameter: a_b : file_target
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--a-b", "aaa"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["a_b"].__class__.__name__, "file_target")
        #
        script = SoS_Script(
            """
parameter: a_b = file_target('file')
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--a-b", "aaa"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["a_b"].__class__.__name__, "file_target")
        #
        script = SoS_Script(
            """
parameter: a_b : sos_targets
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--a-b", "aaa", "bbb"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["a_b"].__class__.__name__, "sos_targets")
        #
        script = SoS_Script(
            """
parameter: a_b = sos_targets('file')
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--a-b", "aaa"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["a_b"].__class__.__name__, "sos_targets")
        #
        #
        #
        #
        script = SoS_Script(
            """
parameter: b : path
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--b", "aaa"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["b"].__class__.__name__, "path")
        #
        script = SoS_Script(
            """
parameter: b = path('file')
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--b", "aaa"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["b"].__class__.__name__, "path")
        #
        script = SoS_Script(
            """
parameter: b : paths
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--b", "aaa", "bbb"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["b"].__class__.__name__, "paths")
        #
        script = SoS_Script(
            """
parameter: b = paths('file')
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--b", "aaa"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["b"].__class__.__name__, "paths")
        #
        script = SoS_Script(
            """
parameter: a_b : path
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--a-b", "aaa"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["a_b"].__class__.__name__, "path")
        #
        script = SoS_Script(
            """
parameter: a_b = path('file')
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--a-b", "aaa"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["a_b"].__class__.__name__, "path")
        #
        script = SoS_Script(
            """
parameter: a_b : paths
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--a-b", "aaa", "bbb"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["a_b"].__class__.__name__, "paths")
        #
        script = SoS_Script(
            """
parameter: a_b = paths('file')
[0]
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--a-b", "aaa"]).run(mode="dryrun")
        self.assertEqual(env.sos_dict["a_b"].__class__.__name__, "paths")
        #

    def test_section_directives(self):
        """Test directives of sections"""
        # cannot be in the global section
        # multi-line OK
        SoS_Script(
            """
[0]
input: 'filename',
    'filename1'

"""
        )
        # An abusive case with multi-line OK, from first column ok
        SoS_Script(
            """
[0]
input: 'filename',
'filename1',
filename4,
opt1=value
output:
    blah

depends:
'something else'
"""
        )
        # option with expression ok
        SoS_Script(
            """
[0]
input: 'filename',  'filename2', opt=value==1

"""
        )
        # unrecognized directive, allowed now
        SoS_Script(
            """
[0]
something: 'filename',  filename2, opt=value==1
"""
        )
        # need commma
        self.assertRaises(
            ParsingError,
            SoS_Script,
            """
[0]
input: 'filename'  filename2
""",
        )
        # can be after action
        SoS_Script(
            """
[0]
func()
input: 'filename',  'filename2', opt=value==1
"""
        )
        # assignments between directives are allowed
        SoS_Script(
            """
[0]
input: 'filename',  'filename2', opt=value==1
a = 'some text'
output: 'filename',  'filename2', opt=value==1
"""
        )
        # can have action between directives
        SoS_Script(
            """
[0]
input: 'filename',  'filename2', opt=value==1
abc
output: 'filename',  'filename2', opt=value==1
"""
        )

    def test_script_format(self):
        """Test putting scripts directly under action"""
        script = SoS_Script(
            """
[0]
input: 'filename',  'filename2', opt=value==1
R:

open('something')
save.put()

"""
        )
        script = SoS_Script(
            """
[0]
input: 'filename',  'filename2', opt=value==1
R: concurrent = True

open('something')
save.put()

"""
        )
        script = SoS_Script(
            """
[0]
input: 'filename',  'filename2', opt=value==1
R: concurrent = True,
    workdir = 'someelse else'

open('something')
save.put()

"""
        )
        # test dedent
        script = SoS_Script(
            """
[0]
python3:
    from sos.runtime import logger
    logger.warning('I am from a dented text')
    if 1:
        logger.warning('Another warning')
"""
        )
        script.workflow()
        # with section head in the script,
        # this will not work even if the embedded
        # python script is perfectly valid.
        self.assertRaises(
            ParsingError,
            SoS_Script,
            '''
[0]
input: 'filename',  'filename2', opt=value==1
python3:

with open('something') as e:
   e.write("""
[section]
""")

''',
        )
        # scripts with section head-like lines
        script = SoS_Script(
            """
[0]
R:
some::function(param)
"""
        )
        script.workflow()
        #
        # script with first-line indent
        #
        script = SoS_Script(
            """
[0]
sh:
  echo "a"

sh('echo "b"')
"""
        )
        script.workflow()
        #
        # script with triple quote and format string
        # #1211
        script = SoS_Script(
            '''
[1]
python3: expand = "${ }"

  ld = ${'100'}
  a =  """doc"""

'''
        )
        wf = script.workflow()
        Base_Executor(wf).run()
        script = SoS_Script(
            '''
[default]
report: expand = "${ }"
  ld_file = ${_input['']:r}
  """ \'\'\'
  {}, ${_output:r},

'''
        )
        wf = script.workflow()
        Base_Executor(wf).run()

    def test_input(self):
        """Test input directive"""
        self.touch(["a.txt", "b.txt", "a.pdf", "a0", "a1"])
        script = SoS_Script(
            """
[0]
files = ['a.txt', 'b.txt']

input: 'a.pdf', files

"""
        )
        wf = script.workflow()
        Base_Executor(wf).run(mode="dryrun")
        #
        # test input types
        script = SoS_Script(
            """
[0:shared={'i':'_input', 'o':'_output'}]
files = (f"a{i}" for i in range(2))
input: {'a.txt', 'b.txt'}, files
output: (f"a{x}" for x in _input)

"""
        )
        wf = script.workflow()
        Base_Executor(wf).run(mode="dryrun")
        self.assertEqual(
            sorted(env.sos_dict["i"]), sos_targets(["a.txt", "a0", "a1", "b.txt"])
        )
        self.assertEqual(
            sorted(env.sos_dict["o"]), sos_targets(["aa.txt", "aa0", "aa1", "ab.txt"])
        )

    def test_group_by(self):
        """Test group_by parameter of step input"""
        # group_by = 'all'
        self.touch(["a{}.txt".format(x) for x in range(15)])
        #
        script = SoS_Script(
            """
[0: shared='executed']

executed = []
input: [f'a{x}.txt' for x in range(1, 5)], group_by='all'

executed.append(_input)

"""
        )
        wf = script.workflow()
        Base_Executor(wf).run(mode="dryrun")
        self.assertEqual(
            env.sos_dict["executed"],
            [sos_targets("a1.txt", "a2.txt", "a3.txt", "a4.txt")],
        )
        self.assertEqual(env.sos_dict["executed"][0].labels, ["0"] * 4)
        # group_by = 'single'
        script = SoS_Script(
            """
[0: shared='executed']

executed = []
input: ['a{}.txt'.format(x) for x in range(1, 5)], group_by='single'

executed.append(_input)

"""
        )
        wf = script.workflow()
        Base_Executor(wf).run(mode="dryrun")
        self.assertEqual(
            env.sos_dict["executed"],
            [
                sos_targets("a1.txt"),
                sos_targets("a2.txt"),
                sos_targets("a3.txt"),
                sos_targets("a4.txt"),
            ],
        )
        # group_by = 'pairs'
        script = SoS_Script(
            """
[0: shared='executed']

executed = []
input: [f'a{x}.txt' for x in range(1, 5)], group_by='pairs'

executed.append(_input)

"""
        )
        wf = script.workflow()
        Base_Executor(wf).run(mode="dryrun")
        self.assertEqual(
            env.sos_dict["executed"],
            [sos_targets("a1.txt", "a3.txt"), sos_targets("a2.txt", "a4.txt")],
        )
        # group_by = 'pairs2'
        script = SoS_Script(
            """
[0: shared='executed']

executed = []
input: [f'a{x}.txt' for x in range(1, 9)], group_by='pairs2'

executed.append(_input)

"""
        )
        wf = script.workflow()
        Base_Executor(wf).run(mode="dryrun")
        self.assertEqual(
            env.sos_dict["executed"],
            [
                sos_targets("a1.txt", "a2.txt", "a5.txt", "a6.txt"),
                sos_targets("a3.txt", "a4.txt", "a7.txt", "a8.txt"),
            ],
        )
        # group_by = 'pairs3'
        script = SoS_Script(
            """
[0: shared='executed']

executed = []
input: [f'a{x}.txt' for x in range(1, 13)], group_by='pairs3'

executed.append(_input)

"""
        )
        wf = script.workflow()
        Base_Executor(wf).run(mode="dryrun")
        self.assertEqual(
            env.sos_dict["executed"],
            [
                sos_targets("a1.txt", "a2.txt", "a3.txt", "a7.txt", "a8.txt", "a9.txt"),
                sos_targets(
                    "a4.txt", "a5.txt", "a6.txt", "a10.txt", "a11.txt", "a12.txt"
                ),
            ],
        )

        # group_by = 'pairwise'
        script = SoS_Script(
            """
[0: shared='executed']

executed = []
input: ['a{}.txt'.format(x) for x in range(1, 5)], group_by='pairwise'

executed.append(_input)

"""
        )
        wf = script.workflow()
        Base_Executor(wf).run(mode="dryrun")
        self.assertEqual(
            env.sos_dict["executed"],
            [
                sos_targets("a1.txt", "a2.txt"),
                sos_targets("a2.txt", "a3.txt"),
                sos_targets("a3.txt", "a4.txt"),
            ],
        )
        # group_by = 'pairwiseN'
        script = SoS_Script(
            """
[0: shared='executed']

executed = []
input: ['a{}.txt'.format(x) for x in range(1, 7)], group_by='pairwise2'

executed.append(_input)

"""
        )
        wf = script.workflow()
        Base_Executor(wf).run(mode="dryrun")
        self.assertEqual(
            env.sos_dict["executed"],
            [
                sos_targets("a1.txt", "a2.txt", "a3.txt", "a4.txt"),
                sos_targets("a3.txt", "a4.txt", "a5.txt", "a6.txt"),
            ],
            f'obtained {env.sos_dict["executed"]}',
        )

        # group_by = 'combinations'
        script = SoS_Script(
            """
[0: shared='executed']

executed = []
input: ['a{}.txt'.format(x) for x in range(1, 5)], group_by='combinations'

executed.append(_input)

"""
        )
        wf = script.workflow()
        Base_Executor(wf).run(mode="dryrun")
        self.assertEqual(
            env.sos_dict["executed"],
            [
                sos_targets("a1.txt", "a2.txt"),
                sos_targets("a1.txt", "a3.txt"),
                sos_targets("a1.txt", "a4.txt"),
                sos_targets("a2.txt", "a3.txt"),
                sos_targets("a2.txt", "a4.txt"),
                sos_targets("a3.txt", "a4.txt"),
            ],
        )

        # group_by = 'combinations3'
        script = SoS_Script(
            """
[0: shared='executed']

executed = []
input: ['a{}.txt'.format(x) for x in range(1, 5)], group_by='combinations3'

executed.append(_input)

"""
        )
        wf = script.workflow()
        Base_Executor(wf).run(mode="dryrun")
        self.assertEqual(
            env.sos_dict["executed"],
            [
                sos_targets(["a1.txt", "a2.txt", "a3.txt"]),
                sos_targets(["a1.txt", "a2.txt", "a4.txt"]),
                sos_targets(["a1.txt", "a3.txt", "a4.txt"]),
                sos_targets(["a2.txt", "a3.txt", "a4.txt"]),
            ],
            f'obtained {env.sos_dict["executed"]}',
        )
        # group_by chunks specified as integers
        script = SoS_Script(
            """
[0: shared='executed']

executed = []
input: ['a{}.txt'.format(x) for x in range(1, 10)], group_by=3

executed.append(_input)

"""
        )
        wf = script.workflow()
        Base_Executor(wf).run(mode="dryrun")
        self.assertEqual(
            env.sos_dict["executed"],
            [
                sos_targets("a1.txt", "a2.txt", "a3.txt"),
                sos_targets("a4.txt", "a5.txt", "a6.txt"),
                sos_targets("a7.txt", "a8.txt", "a9.txt"),
            ],
        )
        # group_by chunks specified as integer strings
        script = SoS_Script(
            """
[0: shared='executed']

executed = []
input: ['a{}.txt'.format(x) for x in range(1, 10)], group_by='3'

executed.append(_input)

"""
        )
        wf = script.workflow()
        Base_Executor(wf).run(mode="dryrun")
        self.assertEqual(
            env.sos_dict["executed"],
            [
                sos_targets("a1.txt", "a2.txt", "a3.txt"),
                sos_targets("a4.txt", "a5.txt", "a6.txt"),
                sos_targets("a7.txt", "a8.txt", "a9.txt"),
            ],
        )
        # number of files should be divisible by group_by
        self.touch(["a{}.txt".format(x) for x in range(1, 10)])
        script = SoS_Script(
            """
[0]

executed = []
input: ['a{}.txt'.format(x) for x in range(1, 10)], group_by=4

executed.append(_input)

"""
        )
        wf = script.workflow()
        Base_Executor(wf).run(mode="dryrun")
        # incorrect value causes an exception
        script = SoS_Script(
            """
[0]

executed = []
input: ['a{}.txt'.format(x) for x in range(1, 10)], group_by='something'

executed.append(_input)

"""
        )
        wf = script.workflow()
        self.assertRaises(Exception, Base_Executor(wf).run, mode="dryrun")
        #
        # group by label
        file_target("c.txt").touch()
        script = SoS_Script(
            """
[A]
output: 'a.txt'
_output.touch()

[B]
input: for_each={'i': range(2)}
output: 'b.txt', 'b1.txt', group_by=1
_output.touch()

[0: shared='executed']
executed = []

input: 'c.txt', output_from(['A', 'B']), group_by='label'

executed.append(_input)

"""
        )
        wf = script.workflow()
        Base_Executor(wf).run(mode="dryrun")
        self.assertEqual(
            env.sos_dict["executed"],
            [
                sos_targets("c.txt"),
                sos_targets("a.txt"),
                sos_targets("b.txt", "b1.txt"),
            ],
        )
        #
        # group_by='pairlabel'
        file_target("c.txt").touch()
        script = SoS_Script(
            """
[A]
output: 'a1.txt', 'a2.txt'
_output.touch()

[B]
output: 'b1.txt', 'b2.txt'
_output.touch()

[0: shared='executed']
executed = []

input: output_from(['A', 'B']), group_by='pairlabel'

executed.append(_input)

"""
        )
        wf = script.workflow()
        Base_Executor(wf).run(mode="dryrun")
        self.assertEqual(
            env.sos_dict["executed"],
            [sos_targets("a1.txt", "b1.txt"), sos_targets("a2.txt", "b2.txt")],
        )
        # group_by='pairlabel3'
        self.touch(["c{}.txt".format(x) for x in range(1, 7)])

        script = SoS_Script(
            """
[A]
output: [f'a{x}.txt' for x in range(1, 7)]
_output.touch()

[B]
output: [f'b{x}.txt' for x in range(1, 7)]
_output.touch()

[0: shared='executed']
executed = []

input: [f'c{x}.txt' for x in range(1, 7)], output_from(['A', 'B']), group_by='pairlabel3'

executed.append(_input)

"""
        )
        wf = script.workflow()
        Base_Executor(wf).run(mode="dryrun")
        self.assertEqual(
            env.sos_dict["executed"],
            [
                sos_targets(
                    "c1.txt",
                    "c2.txt",
                    "c3.txt",
                    "a1.txt",
                    "a2.txt",
                    "a3.txt",
                    "b1.txt",
                    "b2.txt",
                    "b3.txt",
                ),
                sos_targets(
                    "c4.txt",
                    "c5.txt",
                    "c6.txt",
                    "a4.txt",
                    "a5.txt",
                    "a6.txt",
                    "b4.txt",
                    "b5.txt",
                    "b6.txt",
                ),
            ],
        )
        # group_by='pairlabel3'
        self.touch(["c{}.txt".format(x) for x in range(1, 7)])

        script = SoS_Script(
            """
[A]
output: 'a1.txt'
_output.touch()

[B]
output: [f'b{x}.txt' for x in range(1, 7)]
_output.touch()

[0: shared='executed']
executed = []

input: [f'c{x}.txt' for x in range(1, 3)], output_from(['A', 'B']), group_by='pairlabel3'

executed.append(_input)

"""
        )
        wf = script.workflow()
        Base_Executor(wf).run(mode="dryrun")
        self.assertEqual(
            env.sos_dict["executed"],
            [
                sos_targets("c1.txt", "a1.txt", "b1.txt", "b2.txt", "b3.txt"),
                sos_targets("c2.txt", "a1.txt", "b4.txt", "b5.txt", "b6.txt"),
            ],
        )
        #
        # group by function
        file_target("c.txt").touch()
        script = SoS_Script(
            """
[0: shared='executed']
executed = []

def grp(x):
    return  [x[:3], x[3:]]

input: ['a{}.txt'.format(x) for x in range(5)], group_by=grp

executed.append(_input)

"""
        )
        wf = script.workflow()
        Base_Executor(wf).run(mode="dryrun")
        self.assertEqual(
            env.sos_dict["executed"],
            [["a0.txt", "a1.txt", "a2.txt"], ["a3.txt", "a4.txt"]],
        )
        #
        # group by lambda function
        file_target("c.txt").touch()
        script = SoS_Script(
            """
[0: shared='executed']
executed = []

input: ['a{}.txt'.format(x) for x in range(6)], group_by=lambda x: zip(x[:3], x[3:])

executed.append(_input)

"""
        )
        wf = script.workflow()
        Base_Executor(wf).run(mode="dryrun")
        self.assertEqual(
            env.sos_dict["executed"],
            [["a0.txt", "a3.txt"], ["a1.txt", "a4.txt"], ["a2.txt", "a5.txt"]],
        )

    def test_output_group_by(self):
        """Test group_by parameter of step output"""
        # group_by = 'all'
        self.touch(["a{}.txt".format(x) for x in range(4)])
        #
        script = SoS_Script(
            """
[0: shared='executed']

executed = []
input: ['a{}.txt'.format(x) for x in range(4)], group_by=2
output: ['a{}.txt.bak'.format(x) for x in range(4)], group_by=2

executed.append(_output)

"""
        )
        wf = script.workflow()
        Base_Executor(wf).run(mode="dryrun")
        self.assertEqual(
            env.sos_dict["executed"],
            [
                sos_targets("a0.txt.bak", "a1.txt.bak"),
                sos_targets("a2.txt.bak", "a3.txt.bak"),
            ],
        )

    def test_steps_with_step_name(self):
        """Test from steps"""
        script = SoS_Script(
            """
[step_10]

output: 'a.txt'
_output.touch()

[step_20]
input: output_from(step_name.split('_')[0] + '_10')
print(_input)
"""
        )
        wf = script.workflow()
        Base_Executor(wf).run()
        #
        script = SoS_Script(
            """
[step_10]

output: 'a.txt'
_output.touch()

[step_20]
input: output_from(10)
print(_input)
"""
        )
        wf = script.workflow()
        Base_Executor(wf).run()

    def test_section_actions(self):
        """Test actions of sections"""
        SoS_Script(
            """
[0]
func('''
multiline
string''', with_option=1
)
"""
        )
        self.assertRaises(
            ParsingError,
            SoS_Script,
            """
[0]
func(
""",
        )

    def test_longer_code(self):
        """Test definition of classes (with intermediate newlines) in step."""
        script = SoS_Script(
            """# first block

[0: shared='b']
class A:
    def __init__(self):
        pass

    # the newline above should be fine because SoS treat this as
    # regular lines
    def __call__(self):
        return 0

b = A()()

"""
        )
        wf = script.workflow()
        Base_Executor(wf).run()
        self.assertEqual(env.sos_dict["b"], 0)

    def test_combined_workflow(self):
        """Test the creation and execution of combined workfow"""
        script = SoS_Script(
            """
a0 = 0
if 'executed' not in locals():
    executed = []
parameter: a = a0 + 1
[a_1: shared='executed']
executed.append(step_name)
[a_2: shared='executed']
executed.append(step_name)
[a_3: shared='executed']
executed.append(step_name)
[a_4: shared='executed']
executed.append(step_name)
output: 'out_a_4'
[b_1: shared=['executed', 'input_b1']]
executed.append(step_name)
input_b1 = _input
[b_2: shared='executed']
executed.append(step_name)
[b_3: shared='executed']
executed.append(step_name)
[b_4: shared='executed']
executed.append(step_name)
[c: shared='executed']
executed.append(step_name)
[d: shared='executed']
executed.append(step_name)
"""
        )
        wf = script.workflow("a+b")
        Base_Executor(wf).run(mode="dryrun")
        self.assertEqual(
            env.sos_dict["executed"],
            ["a_1", "a_2", "a_3", "a_4", "b_1", "b_2", "b_3", "b_4"],
        )
        self.assertEqual(env.sos_dict["a"], 1)
        self.assertEqual(env.sos_dict["input_b1"], ["out_a_4"])
        #
        env.sos_dict.pop("executed", None)
        wf = script.workflow("a: 1-2 + a:4 + b:3-")
        Base_Executor(wf).run(mode="dryrun")
        self.assertEqual(env.sos_dict["executed"], ["a_1", "a_2", "a_4", "b_3", "b_4"])
        #
        env.sos_dict.pop("executed", None)
        wf = script.workflow("a+c+d")
        Base_Executor(wf).run(mode="dryrun")
        self.assertEqual(
            env.sos_dict["executed"], ["a_1", "a_2", "a_3", "a_4", "c", "d"]
        )

    def test_yaml_config(self):
        """Test config file in yaml format"""
        with open("myconfig.yml", "w") as config:
            config.write(
                """
# Lines beginning with # are skipped when the JSON is parsed, so we can
# put comments into our JSON configuration files
{
    StoreOwner : "John Doe",

    # List of items that we sell
    Fruits: [ "apple", "banana", "pear" ],
    Price: 1.05
}
"""
            )
        with open("config.sos", "w") as sos:
            sos.write(
                """
[0]
print(CONFIG['StoreOwner'])
print(CONFIG.get('StoreOwner', 'something'))
print(CONFIG.get('StoreOwnerSpouse', 'someone else'))
#print(CONFIG.StoreOwner)
"""
            )
        # run the command
        self.assertEqual(
            subprocess.call(
                "sos run config.sos -c myconfig.yml",
                stderr=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
                shell=True,
            ),
            0,
        )
        # now test the value
        script = SoS_Script(filename="config.sos")
        wf = script.workflow()
        Base_Executor(wf, config={"config_file": "myconfig.yml"}).run()
        self.assertEqual(env.sos_dict["CONFIG"]["Price"], 1.05)
        self.assertEqual(env.sos_dict["CONFIG"]["StoreOwner"], "John Doe")
        self.assertEqual(env.sos_dict["CONFIG"]["Fruits"], ["apple", "banana", "pear"])

    def test_var_output(self):
        """Test early appearance of variable output"""
        script = SoS_Script(
            """
[0]
seq = range(3)
input: for_each='seq'
output: f"test{_seq}.txt"
print(_output)
"""
        )
        wf = script.workflow()
        # this does not work before until we make variable output available sooner
        Base_Executor(wf).run(mode="dryrun")

    def test_cell(self):
        """Test ignoring %cell"""
        SoS_Script(
            """
%cell 1
[step ]
a = 1
"""
        )

    def test_overwrite_keyword(self):
        """Test overwrite sos keyword with user defined one."""
        if file_target("a.txt").exists():
            file_target("a.txt").unlink()
        #
        script = SoS_Script(
            """
def run(script):
    pass

[1]
run:
    touch a.txt
"""
        )
        wf = script.workflow()
        Base_Executor(wf).run()
        self.assertFalse(os.path.isfile("a.txt"))
        #
        script = SoS_Script(
            """
parameter: run = 5

[1]
run:
    touch a.txt
"""
        )
        wf = script.workflow()
        # this is ok, see https://github.com/vatlab/SoS/issues/1221
        Base_Executor(wf).run()

    def test_help_message(self):
        """Test help message from ipynb notebook"""
        msg = subprocess.check_output(
            "sos run sample_workflow.ipynb -h", shell=True
        ).decode()
        self.assertFalse("this comment will be included but not shown in help" in msg)
        self.assertTrue(
            msg.count("this comment will become the comment for parameter b"), 1
        )
        self.assertTrue(
            msg.count("this comment will become the comment for parameter d"), 1
        )
        self.assertTrue(msg.count("--c 3 (as int)"), 1)
        self.assertTrue(msg.count("this is a section comment, will be displayed"), 1)
        self.assertFalse("this is a test workflow" in msg)
        self.assertFalse("this is a cell with another kernel" in msg)
        self.assertFalse(
            "this comment will not be included in exported workflow" in msg
        )

    def test_help_on_multi_workflow(self):
        """Test help message of sos file (#985)"""
        with open("test_msg.sos", "w") as script:
            script.write(
                """\
[workflow_a_10,workflow_b]
[workflow_a_20]
[default]
"""
            )
        msg = subprocess.check_output("sos run test_msg.sos -h", shell=True).decode()
        self.assertTrue("workflow_a" in msg)
        self.assertTrue("workflow_b" in msg)
        # when introducing sections
        self.assertTrue("workflow_a_10, workflow_b" in msg)
        self.assertTrue("default" in msg)

    def test_parameter_abbreviation(self):
        """Test potential problem caused by parameter abbreviation #1053"""
        if os.path.isfile("0914.txt"):
            os.remove("0914.txt")
        script = SoS_Script(
            """
[global]
parameter: name = '0914'

[1]
parameter: n = 4
output: f'{name}.txt'
print(_output)
_output.touch()
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--n", "5"]).run()
        self.assertTrue(os.path.isfile("0914.txt"))

    def test_named_input(self):
        """Test named input"""
        for filename in ("a.txt", "b.txt"):
            with open(filename, "w") as out:
                out.write(filename + "\n")
        script = SoS_Script(
            """
[1]
input: {'A': 'a.txt', 'B': 'b.txt'}, group_by='pairlabel'
output: 'c.txt'
assert _input['A'] == [file_target('a.txt')]
assert _input['B'] == [file_target('b.txt')]
with open(_output, 'w') as out:
    out.write(open(_input['A']).read())
    out.write(open(_input['B']).read())
"""
        )
        wf = script.workflow()
        Base_Executor(wf).run()
        self.assertTrue(open("c.txt").read(), "a.txt\nb.txt\n")

    def test_named_output_in_depends(self):
        """Test named_output in depends statement"""
        script = SoS_Script(
            """
[A]
output: A='a.txt'
_output.touch()

[default]
depends: named_output('A')
"""
        )
        wf = script.workflow()
        Base_Executor(wf).run()

    def test_output_from_in_depends(self):
        """Test output_from in depends statement"""
        script = SoS_Script(
            """
[A]
output: A='a.txt'
_output.touch()

[default]
depends: output_from('A')
"""
        )
        wf = script.workflow()
        Base_Executor(wf).run()

    def test_named_output_in_output(self):
        """Test named_output in output statement"""
        script = SoS_Script(
            """
[A]
output: Aa='a.txt'
_output.touch()

[default]
output: named_output('Aa')
"""
        )
        wf = script.workflow()
        self.assertRaises(Exception, Base_Executor(wf).run)

    def test_output_from_in_output(self):
        """Test output_from in output statement"""
        script = SoS_Script(
            """
[Aa]
output: A='a.txt'
_output.touch()

[default]
output: output_from('Aa')
"""
        )
        wf = script.workflow()
        self.assertRaises(Exception, Base_Executor(wf).run)

    def test_sos_step_in_input(self):
        """Test sos_step in input statement"""
        script = SoS_Script(
            """
[A]
output: A='a.txt'
_output.touch()

[default]
input: sos_step('A')
"""
        )
        wf = script.workflow()
        self.assertRaises(Exception, Base_Executor(wf).run)

    def test_sos_variable_in_input(self):
        """Test sos_variable in input statement"""
        script = SoS_Script(
            """
[A]
a=1

[default]
input: sos_variable('a')
"""
        )
        wf = script.workflow()
        self.assertRaises(Exception, Base_Executor(wf).run)

    def test_sos_variable_in_output(self):
        """Test sos_variable in output statement"""
        script = SoS_Script(
            """
[A]
a = 1

[default]
output: sos_variable('a')
"""
        )
        wf = script.workflow()
        self.assertRaises(Exception, Base_Executor(wf).run)

    def test_sos_variable_with_keywordargument(self):
        """Test output_from in output statement"""
        script = SoS_Script(
            """
[A]
a = 1

[default]
depends: sos_variable(var='a')
"""
        )
        wf = script.workflow()
        self.assertRaises(Exception, Base_Executor(wf).run)

    def test_wide_card_step_name(self):
        """test resolving step name with *"""
        script = SoS_Script(
            """
[A_1]

[*_2]
assert step_name == 'A_2', f'step_name is {step_name}, A_2 expected'
"""
        )
        wf = script.workflow()
        Base_Executor(wf).run()

    def test_outfrom_prev_step(self):
        """Test output_from(-1) from output_from """
        script = SoS_Script(
            """
[A_1]
output: 'A_1.txt'
_output.touch()

[A_2]
input: output_from(-1)

[default]
depends: sos_step('A_2')

"""
        )
        wf = script.workflow()
        Base_Executor(wf).run()

    def test_step_from_numeric_step(self):
        """Test sos_step(index) #1209"""
        script = SoS_Script(
            """
[1]

[2]
depends: sos_step('1')
"""
        )
        wf = script.workflow()
        Base_Executor(wf).run()
        #
        script = SoS_Script(
            """
[1]

[2]
depends: sos_step(1)

"""
        )
        wf = script.workflow()
        Base_Executor(wf).run()

    def test_depends_on_step_with_unspecified_input(self):
        for file in ("A_1.txt", "A_2.txt", "A_3.txt", "A_4.txt"):
            if os.path.isfile(file):
                os.remove(file)
        #
        script = SoS_Script(
            """
[A_1]
output: f'{step_name}.txt'
_output.touch()

[A_2]
output: f'{step_name}.txt'
_output.touch()

[A_3]
output: f'{step_name}.txt'
_output.touch()

[A_4]
output: f'{step_name}.txt'
_output.touch()

[default]
depends: sos_step('A_2')
"""
        )
        wf = script.workflow()
        # this should execute A_2 and A_1
        Base_Executor(wf).run()
        self.assertTrue(os.path.isfile("A_1.txt"))
        self.assertTrue(os.path.isfile("A_2.txt"))
        self.assertFalse(os.path.isfile("A_3.txt"))
        self.assertFalse(os.path.isfile("A_4.txt"))
        #
        for file in ("A_1.txt", "A_2.txt", "A_3.txt", "A_4.txt"):
            if os.path.isfile(file):
                os.remove(file)
        #
        script = SoS_Script(
            """
[A_1]
output: f'{step_name}.txt'
_output.touch()

[A_2]
output: f'{step_name}.txt'
_output.touch()

[A_3]
output: f'{step_name}.txt'
_output.touch()

[A_4]
output: f'{step_name}.txt'
_output.touch()

[default]
input: output_from('A_2')
"""
        )
        wf = script.workflow()
        # this should execute A_2 and A_1
        Base_Executor(wf).run()
        self.assertTrue(os.path.isfile("A_1.txt"))
        self.assertTrue(os.path.isfile("A_2.txt"))
        self.assertFalse(os.path.isfile("A_3.txt"))
        self.assertFalse(os.path.isfile("A_4.txt"))
        #
        for file in ("A_1.txt", "A_2.txt", "A_3.txt", "A_4.txt"):
            if os.path.isfile(file):
                os.remove(file)
        #
        script = SoS_Script(
            """
[A_1]
output: f'{step_name}.txt'
_output.touch()

[A_2]
input: None
output: f'{step_name}.txt'
_output.touch()

[A_3]
output: f'{step_name}.txt'
_output.touch()

[A_4]
output: f'{step_name}.txt'
_output.touch()

[default]
input: output_from('A_2')
"""
        )
        wf = script.workflow()
        # this should execute A_2 and A_1
        Base_Executor(wf).run()
        self.assertFalse(os.path.isfile("A_1.txt"))
        self.assertTrue(os.path.isfile("A_2.txt"))
        self.assertFalse(os.path.isfile("A_3.txt"))
        self.assertFalse(os.path.isfile("A_4.txt"))

    def test_output_from_workflow(self):
        """Test output from workflow"""
        #
        #
        for file in ("A_1.txt", "A_2.txt"):
            if os.path.isfile(file):
                os.remove(file)
        #
        script = SoS_Script(
            r"""
[A_1]
output: f'{step_name}.txt'

with open(_output, 'w') as out:
    out.write(step_name)
    out.write('\n')

[A_2]
output: f'{step_name}.txt'

with open(_output, 'w') as out:
    out.write(step_name)
    out.write('\n')

[default]
input: output_from('A')

with open(_input) as content:
    assert content.read() == 'A_2\n'

"""
        )
        wf = script.workflow()
        # this should execute A_2 and A_1
        Base_Executor(wf).run()
        #
        self.assertTrue(os.path.isfile("A_1.txt"))
        self.assertTrue(os.path.isfile("A_2.txt"))

    def test_execute_global_section(self):
        """Global section should be executed only once #1219"""
        script = SoS_Script(
            r"""
[global]
parameter: A=[1, 2]
parameter: B=[]
B.extend(A)

[default]
assert B == [1, 2]
input: for_each=dict(i=range(2))
assert B == [1, 2]
"""
        )
        wf = script.workflow()
        Base_Executor(wf).run()

    def test_para_from_nested_workflow(self):
        """Test passing arguments to nested workflow #1229"""
        script = SoS_Script(
            r"""
[A]
parameter: param = str
print(param)

[B]
print(step_name)

[default]
sos_run('B+A')
"""
        )
        wf = script.workflow()
        Base_Executor(wf, args=["--param", "2"]).run()

    def test_indented_action(self):
        """Test the use of indented action in script format"""
        #
        # not indented
        script = SoS_Script(
            r"""
[1]
if True:
    sh:
echo something

[2]
for i in range(2):
    sh:
echo something
"""
        )
        wf = script.workflow()
        Base_Executor(wf).run()
        # not indented double action
        script = SoS_Script(
            r"""
[1]
if True:
    sh:
echo something

    python:
print(1)
"""
        )
        wf = script.workflow()
        self.assertRaises(Exception, Base_Executor(wf).run)
        # not indented Python structure
        script = SoS_Script(
            r"""
[1]
if True:
    sh:
echo something
else:
    python:
print(1)
"""
        )
        wf = script.workflow()
        self.assertRaises(Exception, Base_Executor(wf).run)
        #  indented
        script = SoS_Script(
            r"""
[1]
if True:
    sh:
        echo something

[2]
for i in range(2):
    sh:
        echo something
"""
        )
        wf = script.workflow()
        Base_Executor(wf).run()
        # indented double action
        script = SoS_Script(
            r"""
[1]
if True:
    sh:
        echo something
    python:
        print(1)
"""
        )
        wf = script.workflow()
        Base_Executor(wf).run()
        # indented Python structure
        script = SoS_Script(
            r"""
[1]
if True:
    sh:
        echo something
else:
    python:
        print(1)
"""
        )
        wf = script.workflow()
        Base_Executor(wf).run()
        # indented, nested structure
        script = SoS_Script(
            r"""
[1]
for i in range(2):
    if True:
        sh:
            echo something
    else:
        python:
            print(1)
"""
        )
        wf = script.workflow()
        Base_Executor(wf).run()
        #
        # wrong nested action
        script = SoS_Script(
            r"""
report:
    name: 'a.txt'
"""
        )
        wf = script.workflow()
        Base_Executor(wf).run()


def test_task_param_var():
    """Test global parameter passed to task parameters #1281"""
    execute_workflow(
        r"""
        [global]
        parameter: job_size = 60

        [1]
        task: trunk_size = job_size
        bash:
                echo 1
        """,
        options={"default_queue": "localhost"},
    )


def test_task_param_var_to_substep():
    """Test global parameter passed to task parameters in substep #1281"""
    execute_workflow(
        r"""
        [global]
        parameter: job_size = 60

        [1]
        input: for_each=dict(i=range(3))

        task: trunk_size = job_size
        bash: expand=True
                echo {i}
        """,
        options={"default_queue": "localhost"},
    )


def test_empty_parameter():
    # parameter: without content #1283
    execute_workflow(
        r"""
        parameter:

        print(1)
        """
    )


def test_sequential_substeps():
    """Test sequential execution of substeps"""
    execute_workflow(
        r"""
        [10: shared='sum']
        sum = 0
        input: for_each=dict(i=range(4)), concurrent=False
        sum += i
        print(f'sum is {sum} at index {_index}')
        """
    )
    assert env.sos_dict["sum"] == 6


def test_limited_concurrency():
    """Set concurrent=INT"""
    execute_workflow(
        r"""
        [10]
        input: for_each=dict(i=range(6)), concurrent=2
        print(i)
        """
    )


def test_concurrent_substep_with_step_import():
    """ Test concurrent substep with step leval import statement #1354"""
    execute_workflow(
        """
        import time
        input: for_each=dict(i=range(2))
        time.sleep(0)
        """
    )


def test_type_hint():
    """We should be able to differentiate type hint and sos action"""
    SoS_Script("""a : list = 5""")
    SoS_Script("""a : list""")
    # action
    SoS_Script("""a : input='filename' """)
    # action
    SoS_Script("""a : expand='${ }' """)


def test_sos_step_in_output():
    """Test sos_step in output statement"""
    with pytest.raises(Exception):
        execute_workflow(
            """
            [A]
            output: A='a.txt'
            _output.touch()

            [default]
            output: sos_step('A')
        """
        )


def test_comments(sample_workflow):
    """Test the use of comments in sos script"""
    # extract workflow from ipynb
    wf = extract_workflow("sample_workflow.ipynb")
    assert "this is a test workflow" not in wf
    assert wf.count("this comment will be included but not shown in help") == 1
    assert wf.count("this comment will become the comment for parameter b") == 1
    assert wf.count("this comment will become the comment for parameter d") == 1
    assert "this is a cell with another kernel" not in wf
    assert "this comment will not be included in exported workflow" not in wf



#!/usr/bin/env python
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import os
import shutil

import pytest

from sos import execute_workflow


@pytest.mark.skipif(not shutil.which('octave'), reason = 'Octave not installed')
def test_octave():
    '''Test action octave'''
    if os.path.isfile('octave_example.txt'):
        os.remove('octave_example.txt')
    execute_workflow(r'''
        [0]
        octave:
            f = fopen("octave_example.txt", "w")
            fprintf(f, "A, B, C, D\n")
            fclose(f)
        ''')
    assert os.path.isfile('octave_example.txt')



#!/usr/bin/env python
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

from sos import execute_workflow


def test_python():
    '''Test python command. This might fail if python3 is the
    default interpreter'''
    execute_workflow(r'''
        [0]
        python: expand='${ }'
        a = {'1': 2}
        print(a)
        ''')

def test_python3():
    execute_workflow(r'''
        [0]
        python3: expand='${ }'
        a = {'1', '2'}
        print(a)
        ''')
